"""
Various wrapper functions/classes which use :mod:`socket` or are strongly tied to functions in this file
which use :mod:`socket`. Part of :mod:`privex.helpers.net` - network related helper code.

**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )

"""
import asyncio
import functools
import socket
import ssl
import time
from ipaddress import ip_network
from typing import Any, Callable, Generator, IO, Iterable, List, Optional, Tuple, Union

import attr

from privex.helpers import settings
from privex.helpers.common import LayeredContext, byteify, empty, empty_if, is_true, stringify, strip_null
from privex.helpers.thread import SafeLoopThread
from privex.helpers.asyncx import await_if_needed, run_coro_thread
from privex.helpers.net.util import generate_http_request, get_ssl_context, ip_is_v6, ip_sock_ver, is_ip
from privex.helpers.net.dns import resolve_ip, resolve_ip_async
from privex.helpers.types import AUTO, AUTO_DETECTED, AnyNum, STRBYTES, T

import logging

log = logging.getLogger(__name__)

__all__ = [
    'AnySocket', 'OpAnySocket', 'SocketContextManager',
    'StopLoopOnMatch', 'SocketWrapper', 'AsyncSocketWrapper', 'send_data_async', 'send_data', 'upload_termbin',
    'upload_termbin_file', 'upload_termbin_async', 'upload_termbin_file_async'
]

AnySocket = Union[ssl.SSLSocket, "socket.socket"]
OpAnySocket = Optional[Union[ssl.SSLSocket, "socket.socket"]]


class SocketContextManager:
    parent_class: Union["SocketWrapper", "AsyncSocketWrapper"]
    
    def __init__(self, parent_class: Union["SocketWrapper", "AsyncSocketWrapper"]):
        self.parent_class = parent_class
    
    def __enter__(self) -> "SocketWrapper":
        log.debug("Entering SocketContextManager")
        self.parent_class.reconnect()
        return self.parent_class
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        log.debug("Exiting SocketContextManager")
        self.parent_class.close()

    async def __aenter__(self) -> "AsyncSocketWrapper":
        log.debug("[async] Entering SocketContextManager")
        await self.parent_class.reconnect()
        return self.parent_class

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        log.debug("[async] Exiting SocketContextManager")
        self.parent_class.close()


class StopLoopOnMatch(Exception):
    def __init__(self, message: str, match: Any = None, compare: str = None, compare_lower: bool = True, **extra):
        self.message = message
        self.match = match
        self.compare = compare
        self.compare_lower = compare_lower
        self.extra = extra
        super().__init__(message)


def _sockwrapper_auto_connect(new_sock: bool = False):
    def _decorator(f):
        @functools.wraps(f)
        def wrapper(self: Union["SocketWrapper"], *args, _sock_tries=0, **kwargs):
            kwargs = dict(kwargs)
            gensock = None
            if kwargs.pop('new_sock', new_sock):
                log.debug("new_sock is true for call to function %s - generating socket to kwarg 'sock'...", f.__name__)
                # kwargs['sock'] = self._select_socket(new_sock=True)
                gensock = SocketTracker.duplicate(self.tracker)
            elif 'sock' in kwargs and kwargs['sock'] not in [None, False, '']:
                gensock = kwargs.pop('sock')
                
            if gensock not in [None, False, '']:
                
                log.debug("'sock' is present for call to function %s...", f.__name__)
                with gensock as sck:
                    log.debug('ensuring socket is open (inside with). now connecting socket.')
                    try:
                        # self.connect(host=kwargs.get('host'), port=kwargs.get('port'), sock=sck)
                        kwargs['sock'] = sck
                    except OSError as e:
                        if 'already connected' in str(e):
                            log.debug('socket already connected. continuing.')
                    log.debug('socket should now be connected. calling function %s', f.__name__)
                    return f(self, *args, **kwargs)
            
            if not self.connected:
                log.debug('instance socket is not connected ( calling function %s )', f.__name__)
    
                if not self.auto_connect:
                    raise ConnectionError(
                        "Would've auto-connected SocketWrapper, but self.auto_connect is False. Please call connect before "
                        "interacting with the socket."
                    )
                if any([empty(self.host, zero=True), empty(self.port, zero=True)]):
                    raise ConnectionError("Tried to auto-connect SocketWrapper, but self.host and/or self.port are empty!")
                log.debug('connecting instance socket ( calling function %s )', f.__name__)
                # self.connect(self.host, self.port)
                self.tracker.connect()
            try:
                _sock_tries += 1
                return f(self, *args, **kwargs)
            except (BrokenPipeError, ConnectionResetError, ConnectionAbortedError) as e:
                if self.error_reconnect and _sock_tries < 3:
                    log.error("The socket appears to have broken. Resetting and trying again. Error was: %s - %s", type(e), str(e))
                    self.tracker.reconnect()
                    return wrapper(self, *args, _sock_tries=_sock_tries, **kwargs)
                raise e
        return wrapper
    return _decorator


def _async_sockwrapper_auto_connect():
    def _decorator(f):
        @functools.wraps(f)
        async def wrapper(self: Union["AsyncSocketWrapper"], *args, _sock_tries=0, **kwargs):
            if not self.tracker.connected:
                if not self.auto_connect:
                    raise ConnectionError(
                        "Would've auto-connected AsyncSocketWrapper, but self.auto_connect is False. Please call connect before "
                        "interacting with the socket."
                    )

                if any([empty(self.host, zero=True), empty(self.port, zero=True)]):
                    raise ConnectionError("Tried to auto-connect AsyncSocketWrapper, but self.host and/or self.port are empty!")
                # await self.connect(self.host, self.port)
                await self.tracker.connect_async()
            try:
                _sock_tries += 1
                return await f(self, *args, **kwargs)
            except (BrokenPipeError, ConnectionResetError, ConnectionAbortedError) as e:
                log.error("The socket appears to have broken. Error was: %s - %s", type(e), str(e))
                if self.error_reconnect and _sock_tries < 3:
                    log.error("Resetting the connection and trying again...")
                    await self.tracker.reconnect_async()
                    return await wrapper(self, *args, _sock_tries=_sock_tries, **kwargs)
                raise e
        return wrapper
    return _decorator


class MockContext:
    def __enter__(self):
        # return self.auto_socket
        return "yes"
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
    
    async def __aenter__(self):
        return "yes"
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass


@attr.s
class SocketTracker:
    """
    Data class used by :class:`.SocketWrapper` / :class:`.AsyncSocketWrapper` for managing sockets
    """
    host: str = attr.ib()
    port: int = attr.ib(converter=int)
    timeout: Union[int, float] = attr.ib(factory=lambda: settings.DEFAULT_SOCKET_TIMEOUT)
    server: bool = attr.ib(default=False, converter=is_true)
    connected: bool = attr.ib(default=False, converter=is_true)
    binded: bool = attr.ib(default=False, converter=is_true)
    listening: bool = attr.ib(default=False, converter=is_true)
    use_ssl: bool = attr.ib(default=False, converter=is_true)
    socket_conf: dict = attr.ib(factory=dict)
    ssl_conf: dict = attr.ib(factory=dict)
    ssl_wrap_conf: dict = attr.ib(factory=dict)
    hostname: str = attr.ib(default=None)
    _ssl_context: ssl.SSLContext = attr.ib(default=None)
    _ssl_socket: ssl.SSLSocket = attr.ib(default=None)
    _loop: asyncio.AbstractEventLoop = attr.ib(default=None)
    _socket: AnySocket = attr.ib(default=None)
    _socket_layer_ctx = attr.ib(default=None)
    
    _host_v4: Optional[str] = attr.ib(default=None)
    _host_v6: Optional[str] = attr.ib(default=None)
    
    _host_v4_resolved: bool = attr.ib(default=False)
    _host_v6_resolved: bool = attr.ib(default=False)
    
    def __attrs_post_init__(self):
        self.hostname = empty_if(self.hostname, self.host, zero=True)
    
    @property
    def family(self) -> int:
        return self.socket_conf.get('family', -1)
    
    @family.setter
    def family(self, value: int):
        self.socket_conf['family'] = value
    
    @property
    def host_v4(self) -> Optional[str]:
        if not self._host_v4_resolved:
            self._host_v4 = resolve_ip(self.host, 'v4')
            self._host_v4_resolved = True
        return self._host_v4

    @property
    def host_v6(self) -> Optional[str]:
        if not self._host_v6_resolved:
            self._host_v6 = resolve_ip(self.host, 'v6')
            self._host_v6_resolved = True
        return self._host_v6
    
    @property
    def socket(self):
        if not self._socket:
            self._socket = socket.socket(**self.socket_conf)
        return self._socket
    
    @socket.setter
    def socket(self, value):
        pass

    @property
    def socket_layer_ctx(self):
        if not self._socket_layer_ctx:
            self._socket_layer_ctx = LayeredContext(MockContext())
        return self._socket_layer_ctx
    
    @socket_layer_ctx.setter
    def socket_layer_ctx(self, value):
        self._socket_layer_ctx = value

    def _make_context(self, **kwargs) -> ssl.SSLContext:
        cnf = {**self.ssl_conf, **kwargs}
        return get_ssl_context(**cnf)
    
    @property
    def ssl_context(self):
        if not self._ssl_context:
            self._ssl_context = self._make_context()
        return self._ssl_context
    
    @ssl_context.setter
    def ssl_context(self, value):
        self._ssl_context = value
    
    @property
    def ssl_socket(self):
        if not self._ssl_socket:
            self._ssl_socket = self.ssl_context.wrap_socket(self.socket, **self.ssl_wrap_conf)
        return self._ssl_socket
    
    @ssl_socket.setter
    def ssl_socket(self, value):
        self._ssl_socket = value

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if not self._loop:
            self._loop = asyncio.get_event_loop()
        return self._loop
    
    @property
    def _auto_socket(self):
        return self.ssl_socket if self.use_ssl else self.socket
    
    @property
    def auto_socket(self) -> AnySocket:
        if not self.connected: self.connect()
        return self._auto_socket

    @property
    def ip_address(self):
        try:
            if empty(self._auto_socket) or empty(self._auto_socket.getpeername()):
                return None
        except Exception as e:
            log.warning("Error while getting peername: %s %s", type(e), str(e))
            return None
        return self._auto_socket.getpeername()[0]
    connected_ip = ip_address
    
    @property
    def connected_port(self):
        try:
            if empty(self._auto_socket) or empty(self._auto_socket.getpeername()):
                return None
        except Exception as e:
            log.warning("Error while getting peername: %s %s", type(e), str(e))
            return None
        return self._auto_socket.getpeername()[1]

    def bind(self, address: Tuple[str, AnyNum] = None, force=False, **kwargs):
        if self.binded and not force:
            return self.auto_socket
        self.auto_socket.bind(address)
        self.binded = True
        return self.auto_socket

    def listen(self, backlog: int = 10, force=False, **kwargs):
        s = self.auto_socket
        if self.listening and not force:
            return s
        self.auto_socket.listen(backlog)
        self.listening = True
        return self.auto_socket
    
    def post_connect(self, sock: AnySocket):
        log.debug("[%s.%s] Connected to host: %s", __name__, self.__class__.__name__, sock.getpeername())
        sock.settimeout(self.timeout)
        return sock
    
    def v6_fallback(self, ex: Exception = None) -> bool:
        ip = self.ip_address
        if self.family == socket.AF_INET6 or (self.family != socket.AF_INET and not empty(ip) and ip_is_v6(ip)):
            if self.host_v4:
                if ex:
                    log.warning(
                        "[%s.%s] Error while using IPv6. Falling back to v4. %s %s",
                        __name__, self.__class__.__name__, type(ex), str(ex)
                    )
                self.family = socket.AF_INET
                return True
        return False
    
    def connect(self, force=False, override_ssl=None, _conn_tries=0) -> AnySocket:
        if not self.connected or force:
            sock = self.socket
            if self.use_ssl and override_ssl in [None, True]:
                sock = self.ssl_socket
            log.debug("[%s.%s] Connecting to host %s on port %s", __name__, self.__class__.__name__, self.host, self.port)

            # log.debug("Connecting to host %s on port %s", self.host, self.port)
            try:
                _conn_tries += 1
                sock.settimeout(self.timeout)
                sock.connect((self.host, self.port))
            except OSError as e:
                if 'already connected' in str(e):
                    log.debug("[%s.%s] Got OSError. Already connected. %s - %s", __name__, self.__class__.__name__, type(e), str(e))
                    self.connected = True
                    return self.post_connect(self.auto_socket)
                if _conn_tries >= 3:
                    raise e
                if not self.v6_fallback(e):
                    log.warning("[%s.%s] Got OSError. Resetting. %s - %s", __name__, self.__class__.__name__, type(e), str(e))
                # self._socket = None
                return self.reconnect(force=True, override_ssl=override_ssl, _conn_tries=_conn_tries)
            # sock.settimeout(self.timeout)
            self.connected = True
            if self.use_ssl:
                self.ssl_socket = sock
            else:
                self.socket = sock
            return self.post_connect(sock)
        sock = self.ssl_socket if self.use_ssl and override_ssl in [None, True] else self.socket
        return self.post_connect(sock)

    def reconnect(self, force=True, override_ssl=None, _conn_tries=0) -> AnySocket:
        if self.connected or force:
            self.disconnect()
        return self.connect(force=True, override_ssl=override_ssl, _conn_tries=_conn_tries)

    async def reconnect_async(self, force=True, override_ssl=None, _conn_tries=0) -> AnySocket:
        if self.connected or force:
            self.disconnect()
        return await self.connect_async(force=True, override_ssl=override_ssl, _conn_tries=_conn_tries)

    async def connect_async(self, force=False, override_ssl=None, _conn_tries=0) -> AnySocket:
        if not self.connected or force:
            sock = self.socket
            if self.use_ssl and override_ssl in [None, True]:
                sock = self.ssl_socket
            log.debug("[async] [%s.%s] Connecting to host %s on port %s (timeout: %s)", __name__, self.__class__.__name__,
                      self.host, self.port, self.timeout)
            try:
                _conn_tries += 1
                sock.settimeout(self.timeout)
                await asyncio.wait_for(self.loop.sock_connect(sock, (self.host, self.port)), self.timeout + 0.1)
            except (OSError, asyncio.TimeoutError) as e:
                if 'already connected' in str(e):
                    log.debug("[%s.%s] Got OSError. Already connected. %s - %s", __name__, self.__class__.__name__, type(e), str(e))
                    self.connected = True
                    return self.post_connect(self.auto_socket)
                if _conn_tries >= 3:
                    raise e
                if not self.v6_fallback(e):
                    log.warning("[%s.%s] Got OSError. Resetting. %s - %s", __name__, self.__class__.__name__, type(e), str(e))
                # self._socket = None
                return await self.reconnect_async(force=True, override_ssl=override_ssl, _conn_tries=_conn_tries)
            # sock.settimeout(self.timeout)
            self.connected = True
        sock = self.ssl_socket if self.use_ssl and override_ssl in [None, True] else self.socket
        return self.post_connect(sock)

    def _shutdown(self, sck):
        try:
            sck.shutdown(socket.SHUT_RDWR)
        except OSError as e:
            if 'not connected' in str(e): return
            log.warning("OSError while shutting down socket: %s %s", type(e), str(e))
        except Exception as e:
            log.warning("Exception while shutting down socket: %s %s", type(e), str(e))

    def _close(self, sck):
        try:
            sck.close()
        except OSError as e:
            log.warning("OSError while closing socket: %s %s", type(e), str(e))
        except Exception as e:
            log.warning("Exception while closing socket: %s %s", type(e), str(e))
    
    def disconnect(self):
        self.connected, self.binded, self.listening = False, False, False
        try:
            log.debug("[%s.%s] Disconnecting socket for host %s on port %s", __name__, self.__class__.__name__, self.host, self.port)
    
            # log.debug()
            if self._socket:
                self._shutdown(self._socket)
                self._close(self._socket)
                # try:
                #     self._socket.shutdown(socket.SHUT_RDWR)
                # except OSError as e:
                #     log.warning("OSError while shutting down socket: %s %s", type(e), str(e))
                # except Exception as e:
                #     log.warning("Exception while shutting down socket: %s %s", type(e), str(e))
                # self._socket.close()
                self._socket = None
            if self._ssl_socket:
                self._shutdown(self._ssl_socket)
                self._close(self._ssl_socket)
                # self._ssl_socket.shutdown(socket.SHUT_RDWR)
                # self._ssl_socket.close()
                self._ssl_socket = None
            return True
        except Exception:
            log.exception("error while closing socket")
            return False
    
    @classmethod
    def duplicate(cls, inst: "SocketTracker", **kwargs) -> "SocketTracker":
        cfg = dict(
            host=inst.host, port=inst.port, timeout=inst.timeout, server=inst.server, use_ssl=inst.use_ssl,
            socket_conf=inst.socket_conf, ssl_conf=inst.ssl_conf, ssl_wrap_conf=inst.ssl_wrap_conf
        )
        cfg = {**cfg, **kwargs}
        return cls(**cfg)
    
    def __enter__(self):
        if self.socket_layer_ctx.virtual_layer == 0:
            self._socket_layer_ctx = None
            if self.connected:
                self.reconnect()
        elif not self.connected: self.connect()
        # return self.auto_socket
        self.socket_layer_ctx.__enter__()
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.socket_layer_ctx.virtual_layer <= 1:
            self.disconnect()
        self.socket_layer_ctx.__exit__(exc_type, exc_val, exc_tb)
        # return self.auto_socket

    async def __aenter__(self):
        if self.socket_layer_ctx.virtual_layer == 0:
            self._socket_layer_ctx = None
            if self.connected:
                await self.reconnect_async()
        elif not self.connected: await self.connect_async()
        # return self.auto_socket
        await self.socket_layer_ctx.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        # self.disconnect()
        if self.socket_layer_ctx.virtual_layer <= 1:
            self.disconnect()
        await self.socket_layer_ctx.__aexit__(exc_type, exc_val, exc_tb)

    def __getattr__(self, item):
        try:
            return super().__getattribute__(item)
        except AttributeError:
            pass
        sock: AnySocket = super().__getattribute__('auto_socket')
        return getattr(sock, item)
    

class SocketWrapper(object):
    """
    A wrapper class to make working with :class:`socket.socket` much simpler.
    
    .. NOTE:: For AsyncIO, use :class:`.AsyncSocketWrapper` instead.
    
    **Features**
    
     * Automatic address family detection - detects whether you have working IPv4 / IPv6, and decides the best way
       to connect to a host, depending on what IP versions that host supports
     * ``Happy Eyeballs`` for IPv6. If something goes wrong with an IPv6 connection, it will fallback to IPv4 if the
       host has it available (i.e. a domain with both ``A`` and ``AAAA`` records)
     * Easy to use SSL, which works with HTTPS and other SSL-secured protocols. Just pass ``use_ssl=True`` in the constructor.
     * Many wrapper methods such as :meth:`.recv_eof`, :meth:`.query`, and :meth:`.http_request` to make working
       with sockets much easier.
      
    
    **Examples**
    
    Send a string of bytes / text to a server, and then read until EOF::
    
        >>> sw = SocketWrapper('icanhazip.org', 80)
        >>> res = sw.query("GET / HTTP/1.1\\nHost: icanhazip.com\\n\\n")
        >>> print(res)
        HTTP/1.1 200 OK
        Server: nginx
        Content-Type: text/plain; charset=UTF-8
        Content-Length: 17
        x-rtfm: Learn about this site at http://bit.ly/icanhazip-faq and do not abuse the service.
        
        2a07:e00::abc
    
    For basic HTTP requests, you can use :meth:`.http_request`, which will automatically send ``Host`` (based on the host you passed),
    and ``User-Agent``. SSL works too, just set ``use_ssl=True``::
    
        >>> sw = SocketWrapper('myip.privex.io', 443, use_ssl=True)
        >>> res = sw.http_request('/?format=json')
        >>> print(res)
        HTTP/1.1 200 OK
        Server: nginx
        Date: Tue, 22 Sep 2020 03:40:48 GMT
        Content-Type: application/json
        Content-Length: 301
        Connection: close
        Access-Control-Allow-Origin: *
        {"error":false,"geo":{"as_name":"Privex Inc.","as_number":210083,"city":"Stockholm","country":"Sweden",
         "country_code":"SE","error":false,"zip":"173 11"},"ip":"2a07:e00::abc","ip_type":"ipv6","ip_valid":true,
         "messages":[], "ua":"Python Privex Helpers ( https://github.com/Privex/python-helpers )"}
    
    Standard low-level sending and receiving data::
        
        >>> sw = SocketWrapper('127.0.0.1', 8888)
        >>> sw.sendall(b"hello world")    # Send the text 'hello world'
        >>> sw.recv(64)                   # read up to 64 bytes of data from the socket
        b"lorem ipsum\n"
        
    """
    DEFAULT_TIMEOUT = empty_if(socket.getdefaulttimeout(), settings.DEFAULT_SOCKET_TIMEOUT, zero=True)

    _context: Optional[ssl.SSLContext]
    _socket: OpAnySocket
    _base_socket: Optional[socket.socket]
    _ssl_socket: Optional[ssl.SSLSocket]
    _layer_context: Optional[LayeredContext]
    _socket_ctx_mgr: SocketContextManager
    # connected: bool
    auto_connect: bool
    auto_listen: bool
    listen_backlog: int
    tracker: SocketTracker
    
    def __init__(
            self, host: str, port: int, server=False, family=-1, type=socket.SOCK_STREAM, proto=-1, fileno=None,
            timeout=DEFAULT_TIMEOUT, use_ssl=False, verify_cert=False, **kwargs
    ):
        self.host, self.port = host, int(port)
        self.server = is_true(server)
        # if self.server and (empty(type) or type == -1):
        #     type = socket.SOCK_STREAM
        # self._socket = kwargs.get('socket', None)
        # self._base_socket = kwargs.get('base_socket', None)
        # self._ssl_socket = kwargs.get('ssl_socket', None)
        _context = kwargs.get('ssl_context', None)
        # self.connected = not (self._socket is None)
        binded, listening = kwargs.get('binded', False), kwargs.get('listening', False)
        check_connectivity = kwargs.get('check_connectivity', settings.CHECK_CONNECTIVITY)
        self.auto_connect = kwargs.get('auto_connect', True)
        self.error_reconnect = kwargs.get('error_reconnect', True)
        self.auto_listen = kwargs.get('auto_listen', True)
        self.listen_backlog = kwargs.get('listen_backlog', 10)
        self.read_timeout = kwargs.get('read_timeout', settings.DEFAULT_READ_TIMEOUT)
        self.send_timeout = kwargs.get('send_timeout', settings.DEFAULT_WRITE_TIMEOUT)
        
        from privex.helpers.net.common import check_v4_async, check_v6_async

        if family == -1 and is_ip(host):
            log.debug("Host '%s' appears to be an IP. Automatically setting address family based on IP.", host)
            family = ip_sock_ver(host)
        
        if family == -1 and check_connectivity:
            host_v4 = resolve_ip(host, 'v4')
            host_v6 = resolve_ip(host, 'v6')
            
            if host_v6 is not None and run_coro_thread(check_v6_async):
                log.debug("Domain %s has one or more IPv6 addresses, and current system appears to have IPv6 connectivity. "
                          "Using domain's IPv6 address: %s", host, host_v6)
                family = socket.AF_INET6
            elif host_v4 is not None and run_coro_thread(check_v4_async):
                log.debug("Domain %s has one or more IPv4 addresses, and current system appears to have IPv4 connectivity. "
                          "Using domain's IPv4 address: %s", host, host_v4)
                family = socket.AF_INET

        # self.use_ssl = use_ssl
        # self.socket_conf = dict(family=family, type=type, proto=proto, fileno=fileno)
        # self.ssl_wrap_conf = dict(
        #     server_hostname=kwargs.get('server_hostname'),
        #     session=kwargs.get('session'),
        #     do_handshake_on_connect=kwargs.get('do_handshake_on_connect', True)
        # )
        # self.ssl_conf = dict(
        #     verify_cert=verify_cert,
        #     check_hostname=kwargs.get('check_hostname'),
        #     verify_mode=kwargs.get('verify_mode')
        # )
        # sck = self._socket if self._socket is not None else socket.socket(**self.socket_conf)
        self.tracker = SocketTracker(
            self.host, self.port,
            timeout=timeout, server=server, binded=binded, connected=kwargs.get('connected', False),
            listening=listening, use_ssl=use_ssl,
            socket_conf=dict(family=family, type=type, proto=proto, fileno=fileno),
            ssl_conf=dict(
                verify_cert=verify_cert,
                check_hostname=kwargs.get('check_hostname'),
                verify_mode=kwargs.get('verify_mode')
            ),
            ssl_wrap_conf=dict(
                server_hostname=kwargs.get('server_hostname'),
                session=kwargs.get('session'),
                do_handshake_on_connect=kwargs.get('do_handshake_on_connect', True)
            ), hostname=kwargs.get('hostname', None)
        )

        _socket = kwargs.get('socket', None)
        _base_socket = kwargs.get('base_socket', None)
        _ssl_socket = kwargs.get('ssl_socket', None)
        
        if _context is not None: self.tracker.ssl_context = _context
        if _socket is not None: self.tracker.socket = _socket
        if _base_socket is not None: self.tracker.socket = _base_socket
        if _ssl_socket is not None: self.tracker.ssl_socket = _ssl_socket
        
        # self._timeout = float(timeout)
        self._layer_context = None
        self._socket_ctx_mgr = SocketContextManager(self)
        # if use_ssl:
        #     ctx = get_ssl_context(**ssl_params)
        #     s = ctx.wrap_socket(
        #         server_hostname=kwargs.get('server_hostname'),
        #         session=kwargs.get('session'),
        #         do_handshake_on_connect=kwargs.get('do_handshake_on_connect', True),
        #     )
    
    @property
    def ssl_conf(self) -> dict:
        return self.tracker.ssl_conf
    
    @ssl_conf.setter
    def ssl_conf(self, value):
        self.tracker.ssl_conf = value

    @property
    def ssl_wrap_conf(self) -> dict:
        return self.tracker.ssl_wrap_conf

    @ssl_wrap_conf.setter
    def ssl_wrap_conf(self, value):
        self.tracker.ssl_wrap_conf = value

    @property
    def socket_conf(self) -> dict:
        return self.tracker.socket_conf

    @socket_conf.setter
    def socket_conf(self, value):
        self.tracker.socket_conf = value
    
    @property
    def timeout(self):
        return self.tracker.timeout
    
    @property
    def _auto_socket(self):
        return self.tracker._auto_socket
    
    @timeout.setter
    def timeout(self, value):
        self.socket.settimeout(value)
        self.tracker.timeout = value
        # self.base_socket.settimeout(value)
        # if self._socket:
        #     self._socket.settimeout(value)
        # self._timeout = value
    
    def _make_context(self, **kwargs) -> ssl.SSLContext:
        cnf = {**self.ssl_conf, **kwargs}
        return get_ssl_context(**cnf)
    
    def _make_socket(self, **kwargs) -> socket.socket:
        cnf = {**self.socket_conf, **kwargs}
        # if self.server:
        #     if 'family' in cnf: del cnf['family']
        #     if 'type' in cnf: del cnf['type']
        #     if 'proto' in cnf: del cnf['proto']
        #     if 'fileno' in cnf: del cnf['fileno']
        #     log.info("socket host: %s || port: %s", self.host, self.port)
        #     log.info("socket extra config: %s", cnf)
        #     return socket.create_server((self.host, self.port), **cnf)
        return socket.socket(**cnf)

    def _ssl_wrap_socket(self, sock: socket.socket = None, ctx: ssl.SSLContext = None, **kwargs) -> ssl.SSLSocket:
        cnf = {**self.ssl_wrap_conf, **kwargs}
        ctx = empty_if(ctx, self.context, itr=True, zero=True)
        sock = empty_if(sock, self.base_socket, itr=True, zero=True)
        return ctx.wrap_socket(sock, **cnf)
    
    def _select_socket(self, new_sock=False, **kwargs) -> Union[ssl.SSLSocket, "socket.socket"]:
        if new_sock:
            sock = self._make_socket()
            if kwargs.get('use_ssl', self.use_ssl):
                sock = self._ssl_wrap_socket(sock, **kwargs)
            return sock
        if self.use_ssl:
            return self._ssl_wrap_socket(**kwargs)
        return self.base_socket
    
    @property
    def hostname(self):
        return self.tracker.hostname
    
    @hostname.setter
    def hostname(self, value):
        self.tracker.hostname = value
    
    @property
    def context(self) -> ssl.SSLContext:
        # if not self._context:
        #     self._context = self._make_context()
        # return self._context
        if not self.tracker.ssl_context:
            self.tracker.ssl_context = self._make_context()
        return self.tracker.ssl_context
    
    ssl_context = context

    @property
    def base_socket(self) -> socket.socket:
        if not self.tracker.socket:
            self.tracker.socket = self._make_socket()
        return self.tracker.socket

    @base_socket.setter
    def base_socket(self, value: socket.socket):
        self.tracker.socket = value

    @property
    def socket(self) -> AnySocket:
        # if not self._socket:
        #     self._socket = self._select_socket()
        # if not self.server: self._socket.settimeout(self.timeout)
        return self.tracker.auto_socket
    
    @socket.setter
    def socket(self, value: AnySocket):
        if self.tracker.use_ssl:
            self.tracker.ssl_socket = value
        else:
            self.tracker.socket = value
        # self._socket = value
    
    @property
    def connected(self):
        return self.tracker.connected
    
    # @connected.setter
    # def connected(self, value):
    #     self.tracker.connected = value
    
    def _connect_sanity(self, host, port, sock: OpAnySocket = None, **kwargs):
        port = int(port)
        sck = self.socket if sock is None else sock

        if sock is None and self.connected and self.socket is not None:
            if host != self.host or port != int(self.port):
                log.debug(f"Already connected, but {self.__class__.__name__}.connect called with different host/port than stored. "
                          f"Trigerring a reconnect.")
                return self.reconnect(host, port, sock=sck)
            log.debug(f"Already connected, {self.__class__.__name__}.connect called with same details as previously. "
                      f"Returning existing socket.")
            return sck
        if empty(port, True, True):
            raise ValueError(f"{self.__class__.__name__}.connect requires a port. Either connect(host, port) or connect( (host,port) )")
        return True
        
    def _connect(self, host: str, port: AnyNum, sock: OpAnySocket = None, **kwargs) -> AnySocket:
        port = int(port)
        sck = self.tracker if sock is None else sock
        if self.server:
            log.debug("Binding to host '%s' on port %s", host, port)
            self.bind(host, port, sock=sock)
            log.debug("Successfully binded to host '%s' on port %s", host, port)
            if self.auto_listen:
                log.debug("Auto-listen is enabled. Calling %s.listen(%s)", self.__class__.__name__, self.listen_backlog)
                self.listen(self.listen_backlog, sock=sock)
                log.debug("%s is now listening on host(s) '%s' on port %s", self.__class__.__name__, host, port)
            # if sock is None: self.host, self.port, self.connected = host, port, True
            return sck

        log.debug("[%s.%s] Connecting to host %s on port %s", self.__class__.__name__, __name__, host, port)
        sck.connect((host, port))
        
        # if sock is None: self.host, self.port, self.connected = host, port, True
        return sck
    
    def _get_addr(self, host: Union[str, Tuple[str, AnyNum]] = None, port: AnyNum = None) -> Tuple[str, int]:
        csn = self.__class__.__name__
    
        if host is None:
            if self.host is None: raise ValueError(f"No host specified to {csn}.reconnect(host, port) - and no host in {csn}.host")
            host = self.host
        if port is None:
            if self.port is None: raise ValueError(f"No port specified to {csn}.connect(host, port) - and no port in {csn}.port")
            port = self.port
        if isinstance(host, (list, set)): host = tuple(host)
        if isinstance(host, tuple): host, port = host
        
        return host, int(port)
    
    def bind(self, host: Union[str, Tuple[str, AnyNum]] = None, port: AnyNum = None, sock: OpAnySocket = None, **kwargs):
        sck = self.socket if sock is None else sock
        if sock is None and self.binded:
            return
        sck.bind(self._get_addr(host, port))
        if sock is None: self.binded = True
        return True
    
    def connect(self, host: Union[str, Tuple[str, AnyNum]] = None, port: AnyNum = None, sock: OpAnySocket = None, **kwargs) -> AnySocket:
        # csn = self.__class__.__name__
        #
        # if host is None:
        #     if self.host is None: raise ValueError(f"No host specified to {csn}.reconnect(host, port) - and no host in {csn}.host")
        #     host = self.host
        # if port is None:
        #     if self.port is None: raise ValueError(f"No port specified to {csn}.connect(host, port) - and no port in {csn}.port")
        #     port = self.port
        # if isinstance(host, (list, set)): host = tuple(host)
        # if isinstance(host, tuple): host, port = host

        host, port = self._get_addr(host, port)
        sanity = self._connect_sanity(host, port, sock=sock)
        if sanity is not True: return sanity
        return self._connect(host, port, sock=sock)
    
    def reconnect(self, host: Union[str, Tuple[str, AnyNum]] = None, port: AnyNum = None, sock: OpAnySocket = None, **kwargs):
        csn = self.__class__.__name__

        # self.close(sock=sock)
        if host is None:
            if port is not None:
                if self.host is None:
                    raise ValueError(f"No host specified to {csn}.reconnect(host, port) - and no host in {csn}.host")
                # return self.connect(self.host, port, sock=sock, **kwargs)
                host = self.host
                # self.tracker.host, self.tracker.port = self.host, port
                # self.tracker.reconnect()
                # return self.tracker
            # if all([self.host is not None, self.port is not None]):
                # return self.connect(self.host, self.port, sock=sock, **kwargs)
                # self.tracker.host, self.tracker.port = self.host,

                # self.tracker.connect()
                # return self.tracker
        elif port is None:
            port = self.port
            # self.tracker.host, self.tracker.port = host, port
        # else:
        self.tracker.host, self.tracker.port = host, port
    
        # return self.connect(host, port, sock=sock, **kwargs)
        self.tracker.reconnect()
        return self.tracker
        # return self.connect(host, self.port, sock=sock, **kwargs)
    
    def listen(self, backlog=10, sock: OpAnySocket = None, **kwargs):
        if self.listening:
            return True
        (self.socket if sock is None else sock).listen(backlog)
        if sock is None: self.listening = True
        return True
    
    @_sockwrapper_auto_connect()
    def accept(self, sock: OpAnySocket = None, **kwargs) -> Tuple[AnySocket, Tuple[str, int]]:
        return (self.socket if sock is None else sock).accept()
    
    @_sockwrapper_auto_connect()
    def settimeout(self, value, sock: OpAnySocket = None, **kwargs):
        return (self.socket if sock is None else sock).settimeout(value)
    
    def close(self, sock: OpAnySocket = None):
        log.debug("Closing socket connection to host: %s || port: %s", self.host, self.port)
        if sock is not None:
            log.debug(" !! sock was specified. only closing sock.")
            try:
                sock.close()
                log.debug("Closed sock.")
            except Exception:
                log.exception("error while closing sock")
            return
        self.tracker.disconnect()
        # try:
        #     if self._socket is not None:
        #         log.debug("closing self.socket")
        #         self.socket.close()
        # except Exception:
        #     log.exception("error while closing self.socket")
        # try:
        #     if self._base_socket is not None:
        #         log.debug("closing self.base_socket")
        #         self.base_socket.close()
        # except Exception:
        #     log.exception("error while closing self.base_socket")
        #
        # try:
        #     if self._ssl_socket is not None:
        #         self._ssl_socket.close()
        #         log.debug("closing self._ssl_socket")
        # except Exception:
        #     log.exception("error while closing self._ssl_socket")
        # self.connected = False
        # log.debug("setting socket instance attributes to None")
        # self._socket, self._ssl_socket, self._base_socket = None, None, None
    
    @_sockwrapper_auto_connect()
    def recv(self, bufsize: int, flags: int = None, sock: OpAnySocket = None, **kwargs) -> bytes:
        if flags is None: return (self.socket if sock is None else sock).recv(bufsize)
        return (self.socket if sock is None else sock).recv(bufsize, flags)

    @_sockwrapper_auto_connect()
    def recvfrom(self, bufsize: int, flags: int = None, sock: OpAnySocket = None, **kwargs) -> Tuple[bytes, Any]:
        if flags is None: return (self.socket if sock is None else sock).recvfrom(bufsize)
        return (self.socket if sock is None else sock).recvfrom(bufsize, flags)

    @_sockwrapper_auto_connect()
    def recvmsg(
            self, bufsize: int, ancbufsize:int = None, flags: int = None, sock: OpAnySocket = None, **kwargs
    ) -> Tuple[bytes, List[Tuple[int, int, bytes]], int, Any]:
        args = [bufsize]
        if ancbufsize is not None: args.append(ancbufsize)
        if flags is not None: args.append(flags)
        return (self.socket if sock is None else sock).recvmsg(*args)

    @_sockwrapper_auto_connect()
    def read_eof(
                self, bufsize: int = 256, eof_timeout: AnyNum = 120, flags: int = None, timeout_fail=False, strip=True,
                conv: Optional[Callable[[Union[bytes, str]], T]] = stringify, sock: OpAnySocket = None, **kwargs
            ) -> Union[bytes, str, T]:
        strip_func = kwargs.get('strip_func', lambda d: strip_null(d, conv=conv))
        data = b''
        total_time = 0.0
        
        while True:
            st_time = time.time()
            chunk = self.recv(bufsize, flags, sock=sock)
            if not chunk:
                log.debug("Finished reading until EOF")
                break
            e_time = time.time()
            total_time += (e_time - st_time)
            data += chunk
            if total_time > eof_timeout:
                log.error("Giving up, spent over %f seconds (%f) reading until EOF for host %s", eof_timeout, total_time, self.host)
                if timeout_fail:
                    raise TimeoutError(f"Giving up, spent over {eof_timeout} seconds ({total_time}) reading until EOF for host {self.host}")
                break
        
        return strip_func(data) if strip else data

    @_sockwrapper_auto_connect()
    def shutdown(self, how: int = None, sock: OpAnySocket = None, **kwargs):
        how = empty_if(how, socket.SHUT_RDWR, itr=True)
        return (self.socket if sock is None else sock).shutdown(how)

    @_sockwrapper_auto_connect()
    def send(self, data: Union[str, bytes], flags: int = None, sock: OpAnySocket = None, **kwargs):
        a = [byteify(data)]
        if not empty(flags): a.append(flags)
        return (self.socket if sock is None else sock).send(*a)

    @_sockwrapper_auto_connect()
    def sendall(self, data: Union[str, bytes], flags: int = None, sock: OpAnySocket = None, **kwargs):
        a = [byteify(data)]
        if not empty(flags): a.append(flags)
        return (self.socket if sock is None else sock).sendall(*a)

    @_sockwrapper_auto_connect()
    def sendto(self, data: Union[str, bytes], *args, sock: OpAnySocket = None, **kwargs):
        return (self.socket if sock is None else sock).sendto(byteify(data), *args, **kwargs)

    @_sockwrapper_auto_connect()
    def send_chunks(self, gen: Union[Iterable, Generator], flags: int = None, sock: OpAnySocket = None, **kwargs):
        results = []
        for c in gen:
            results.append(self.send(c, flags, sock=sock, **kwargs))
        return results

    # @_sockwrapper_auto_connect()
    # def query(self, data: Union[str, bytes], bufsize: int = 32, eof_timeout=30, **kwargs):
    #     timeout_fail, send_flags = kwargs.get('timeout_fail'), kwargs.get('send_flags', kwargs.get('flags', None))
    #     recv_flags = kwargs.get('recv_flags', kwargs.get('flags', None))
    #     log.debug(" >> Sending %s bytes to %s:%s", len(data), self.host, self.port)
    #     self.sendall(byteify(data), flags=send_flags)
    #     log.debug(" >> Reading %s bytes per chunk from %s:%s", bufsize, self.host, self.port)
    #     return self.read_eof(bufsize, eof_timeout=eof_timeout, flags=recv_flags, timeout_fail=timeout_fail)

    # @_sockwrapper_auto_connect()
    # def http_request(
    #         self, url="/", host=AUTO_DETECTED, method="GET", user_agent=DEFAULT_USER_AGENT, extra_data: Union[STRBYTES, List[str]] = None,
    #         body: STRBYTES = None, eof_timeout=30, **kwargs
    # ) -> Union[bytes, Awaitable[bytes]]:
    #     bufsize, flags, timeout_fail = kwargs.pop('bufsize', 256), kwargs.pop('flags', None), kwargs.pop('timeout_fail', False)
    #     data = self._http_request(url, host=host, method=method, user_agent=user_agent, extra=extra_data, body=body, **kwargs)
    #     self.sendall(data, flags=flags)
    #     return self.read_eof(bufsize, eof_timeout=eof_timeout, flags=flags, timeout_fail=timeout_fail)

    def _http_request(self, url, host: str, method: str, user_agent: str = settings.DEFAULT_USER_AGENT, extra=None, **kwargs) -> bytes:
        host = self.hostname if host == AUTO_DETECTED else host
        return generate_http_request(url, host, method=method, user_agent=user_agent, extra_data=extra, **kwargs)

    @_sockwrapper_auto_connect()
    def query(self, data: Union[str, bytes], bufsize: int = 32, eof_timeout=30, sock: OpAnySocket = None, **kwargs):
        timeout_fail, send_flags = kwargs.pop('timeout_fail', False), kwargs.pop('send_flags', kwargs.get('flags', None))
        recv_flags = kwargs.pop('recv_flags', kwargs.pop('flags', None))
        log.debug(" >> Sending %s bytes to %s:%s", len(data), self.host, self.port)
        self.sendall(byteify(data), flags=send_flags, sock=sock)
        log.debug(" >> Reading %s bytes per chunk from %s:%s", bufsize, self.host, self.port)
        return self.read_eof(bufsize, eof_timeout=eof_timeout, flags=recv_flags, timeout_fail=timeout_fail, sock=sock, **kwargs)

    @_sockwrapper_auto_connect()
    def http_request(
                self, url="/", host=AUTO_DETECTED, method="GET", user_agent=settings.DEFAULT_USER_AGENT,
                extra_data: Union[STRBYTES, List[str]] = None, body: STRBYTES = None, eof_timeout=30, bufsize: int = 256,
                conv: Optional[Callable[[Union[bytes, str]], T]] = stringify, sock: OpAnySocket = None, **kwargs
            ) -> Union[str, bytes, T]:
        
        data = self._http_request(url, host=host, method=method, user_agent=user_agent, extra=extra_data, body=body, **kwargs)
        kargs = dict(data=data, bufsize=bufsize, eof_timeout=eof_timeout, timeout_fail=kwargs.get('timeout_fail', False), conv=conv,
                     sock=sock, **kwargs)
        if sock is not None: return self.query(**kargs)
        # with self:
        with self.tracker:
            return self.query(**kargs)

    @_sockwrapper_auto_connect()
    def setblocking(self, flag: bool, sock: OpAnySocket = None, **kwargs):
        return (self.socket if sock is None else sock).setblocking(flag)
    
    def handle_connection(
            self, sock: AnySocket, addr: Tuple[str, int], callback: Callable[["SocketWrapper", Tuple[str, int]], Any],
            stop_return: Union[str, bytes] = None,
            **kwargs
    ):
        stop_compare, stop_compare_lower = kwargs.get('stop_compare', 'equal'), kwargs.get('stop_compare_lower', True)
        if stop_return is not None: stop_return = stringify(stop_return)
        log.info("NEW CONNECTION: %s || %s", sock, addr)
        log.info("Running callback: %s(%s, %s)\n", callback.__name__, sock, addr)
        orig_cres = callback(self.from_socket(sock), addr)
        cres = stringify(orig_cres)
        log.info("Callback return data: %s\n\n", cres)
        if stop_return is not None:
            if stop_compare_lower: stop_return, cres = stop_return.lower(), cres.lower()
            
            if stop_compare.lower() in ['in', 'contain', 'contains', 'contained', 'within', 'inside']:
                if stop_return in cres or strip_null(stop_return) in strip_null(cres):
                    raise StopLoopOnMatch("Matched stop_return. Parent should stop loop.", cres, stop_compare, stop_compare_lower)
            
            if cres == stop_return or strip_null(cres) == strip_null(stop_return):
                raise StopLoopOnMatch("Matched stop_return. Parent should stop loop.", cres, stop_compare, stop_compare_lower)
        return orig_cres

    @_sockwrapper_auto_connect()
    def on_connect(
            self, callback: Callable[["SocketWrapper", Tuple[str, int]], Any], timeout: AnyNum = None,
            stop_return: Union[str, bytes] = None, **kwargs
    ):
        if not self.server:
            raise ValueError("This SocketWrapper has 'server' set to False. Can't handle incoming connections.")
        if not self.binded: self.bind()
        if not self.listening: self.listen(self.listen_backlog)
        stop_return_match = None
        
        while self.connected and stop_return_match is None:
            log.info("Waiting for incoming connection ( %s:%s || %s ) ...", self.host, self.port, self.socket.getsockname())
            sock, addr = self.accept()
            try:
                self.handle_connection(sock, addr, callback, stop_return, **kwargs)
            except StopLoopOnMatch as e:
                log.info(" !!! Stopping on_connect as 'stop_return' has been matched: %s", stop_return)
                log.info(" !!! The matching message was: %s", e.match)
                break
            
        # if stop_return_match is not None:
        #     log.info(" !!! Stopping on_connect as 'stop_return' has been matched: %s", stop_return)
        #     log.info(" !!! The matching message was: %s", stop_return_match)

        log.info(" !!! Disconnected. Stopping on_connect.")

    class SocketWrapperThread(SafeLoopThread):
        def __init__(self, *args, parent_instance: "SocketWrapper", callback, stop_return, conn_kwargs: dict = None, **kwargs):
            kwargs = dict(kwargs)
            self.parent_instance = parent_instance
            self.callback = callback
            self.conn_kwargs = empty_if(conn_kwargs, {}, itr=True)
            self.stop_return = stop_return
            self.stop_compare = kwargs.pop('stop_compare', 'equal')
            self.stop_compare_lower = kwargs.pop('stop_compare_lower', True)
            super().__init__(*args, **kwargs)
    
        def loop(self):
            pi = self.parent_instance
            log.info("Waiting for incoming connection ( %s:%s || %s ) ...", pi.host, pi.port, pi.socket.getsockname())
            sock, addr = pi.accept()
            try:
                pi.handle_connection(
                    sock, addr, self.callback, self.stop_return,
                    stop_compare=self.stop_compare, stop_compare_lower=self.stop_compare_lower, **self.conn_kwargs
                )
            except StopLoopOnMatch as e:
                log.info(" !!! Stopping on_connect as 'stop_return' has been matched: %s", self.stop_return)
                log.info(" !!! The matching message was: %s", e.match)
                self.emit_stop()
        
        def run(self):
            self.parent_instance.reconnect()
            return super().run()
        
    def on_connect_thread(
            self, callback: Callable[["SocketWrapper", Tuple[str, int]], Any], timeout: AnyNum = None,
            stop_return: Union[str, bytes] = None, daemon=True, auto_start=True, **kwargs
    ) -> SocketWrapperThread:
        t = self.SocketWrapperThread(parent_instance=self, callback=callback, stop_return=stop_return, **kwargs)
        t.setDaemon(daemon)
        if auto_start:
            t.start()
        return t

    @classmethod
    def from_socket(cls, sock: AnySocket, server=False, **kwargs) -> Union["SocketWrapper", "AsyncSocketWrapper"]:
        sock_host, sock_port = sock.getsockname()
        cfg = dict(
            family=sock.family, proto=sock.proto, type=sock.type, fileno=sock.fileno(),
            host=sock_host, port=sock_port, server=server, socket=sock, base_socket=sock
        )
        cfg = {**cfg, **kwargs}
        return cls(**cfg)
    
    def __getattribute__(self, item):
        try:
            return super().__getattribute__(item)
        except AttributeError:
            pass
        sock: AnySocket = super().__getattribute__('socket')
        return getattr(sock, item)
    
    def __enter__(self):
        # if not self._socket_ctx_mgr:
        #     self._socket_ctx_mgr = SocketContextManager(self)
        # if not self._layer_context:
        #     self._layer_context = LayeredContext(self._socket_ctx_mgr, max_layers=1)
        # return self._layer_context.__enter__()
        self.tracker.__enter__()
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        # return self.tracker.__aexit__()
    
        # return self._layer_context.__exit__(exc_type, exc_val, exc_tb)
        return self.tracker.__exit__(exc_type, exc_val, exc_tb)


class AsyncSocketWrapper(SocketWrapper):
    """
    
        >>> from privex.helpers import AsyncSocketWrapper
        >>> sw = AsyncSocketWrapper('termbin.com', 9999)
        >>> url = await sw.query("HELLO world\\n\\nThis is a test\\nusing async sockets\\n\\nwith Python")
        'https://termbin.com/lsd93'
        >>> url = await sw.read_eof()
    """
    _loop: Optional[asyncio.AbstractEventLoop]
    DEFAULT_TIMEOUT = empty_if(socket.getdefaulttimeout(), settings.DEFAULT_SOCKET_TIMEOUT, zero=True)

    def __init__(
            self, host: str, port: int, server=False, family=-1, type=socket.SOCK_STREAM, proto=-1, fileno=None,
            timeout=DEFAULT_TIMEOUT, use_ssl=False, verify_cert=False, loop=None, **kwargs
    ):
        self._loop = loop
        super().__init__(
            host=host, port=port, server=server, family=family, type=type, proto=proto, fileno=fileno, timeout=timeout,
            use_ssl=use_ssl, verify_cert=verify_cert, **kwargs
        )
    
    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if not self._loop:
            self._loop = asyncio.get_event_loop()
        return self._loop
    
    async def _connect(self, host: str, port: AnyNum, sock: OpAnySocket = None, **kwargs) -> AnySocket:
        port = int(port)
        # sck = self.socket if sock is None else sock

        if self.server:
            log.debug("Binding to host '%s' on port %s", host, port)
            self.bind(host, port, sock=sock)
            log.debug("Successfully binded to host '%s' on port %s", host, port)
            if self.auto_listen:
                log.debug("Auto-listen is enabled. Calling %s.listen(%s)", self.__class__.__name__, self.listen_backlog)
                self.listen(self.listen_backlog, sock=sock)
                log.debug("%s is now listening on host(s) '%s' on port %s", self.__class__.__name__, host, port)
            # if sock is None: self.host, self.port, self.connected = host, port, True
            return self.socket
        log.debug("Connecting to host %s on port %s", host, port)
        if sock:
            await self.loop.sock_connect(sock, (host, port))
            return sock
        await self.tracker.connect_async()
        return self.tracker.auto_socket
        # self.loop.soc
        # if sock is None: self.host, self.port, self.connected = host, port, True
    
    async def connect(
            self, host: Union[str, Tuple[str, AnyNum]] = None, port: AnyNum = None, sock: OpAnySocket = None, **kwargs) -> AnySocket:
        
        host, port = self._get_addr(host, port)
        sanity = await self._connect_sanity(host, port)
        if sanity is not True: return sanity
        return await self._connect(host, port, sock=sock)

    async def _connect_sanity(self, host, port, sock: OpAnySocket = None, **kwargs):
        port = int(port)
        # sock = self._auto_socket if sock is None else sock

        if sock is not None or (self.connected and self._auto_socket is not None):
            if host != self.host or port != int(self.port):
                log.debug(f"Already connected, but {self.__class__.__name__}.connect called with different host/port than stored. "
                          f"Trigerring a reconnect.")
                return await self.reconnect(host, port, sock=sock)
            log.debug(f"Already connected, {self.__class__.__name__}.connect called with same details as previously. "
                      f"Returning existing socket.")
            return sock
        if empty(port, True, True):
            raise ValueError(f"{self.__class__.__name__}.connect requires a port. Either connect(host, port) or connect( (host,port) )")
        return True

    async def reconnect(self, host: Union[str, Tuple[str, AnyNum]] = None, port: AnyNum = None, sock: OpAnySocket = None, **kwargs):
        csn = self.__class__.__name__
        self.close()
        if host is None:
            if port is not None:
                if self.host is None:
                    raise ValueError(f"No host specified to {csn}.reconnect(host, port) - and no host in {csn}.host")
                return await self.connect(self.host, port, sock=sock)
            if all([self.host is not None, self.port is not None]):
                return await self.connect(self.host, self.port, sock=sock)
        if port is not None:
            return await self.connect(host, port, sock=sock)
        return await self.connect(host, self.port, sock=sock)

    @_async_sockwrapper_auto_connect()
    async def read_eof(
            self, bufsize: int = 256, eof_timeout: AnyNum = 120, flags: int = None, timeout_fail=False, strip=True,
            conv: Optional[Callable[[Union[bytes, str]], T]] = stringify, sock: OpAnySocket = None, **kwargs
    ) -> Union[str, bytes, T]:
        strip_func = kwargs.get('strip_func', lambda d: strip_null(d, conv=conv))
        data, total_time = b'', 0.0
    
        while True:
            st_time = time.time()
            chunk = await self.recv(bufsize, flags, timeout=kwargs.get('read_timeout', AUTO), sock=sock)
            if not chunk:
                log.debug("Finished reading until EOF")
                break
            e_time = time.time()
            total_time += (e_time - st_time)
            data += chunk
            if not empty(eof_timeout, True) and total_time > eof_timeout:
                log.error("Giving up, spent over %f seconds (%f) reading until EOF for host %s", eof_timeout, total_time, self.host)
                if timeout_fail:
                    raise TimeoutError(f"Giving up, spent over {eof_timeout} seconds ({total_time}) reading until EOF for host {self.host}")
                break
    
        return strip_func(data) if strip else data

    @_async_sockwrapper_auto_connect()
    async def recv(self, bufsize: int, flags: int = None, sock: OpAnySocket = None, timeout: Union[float, int] = AUTO, **kwargs) -> bytes:
        timeout, sck = self.read_timeout if timeout is AUTO else timeout, self.socket if sock is None else sock
        if timeout not in [None, False]:
            return await asyncio.wait_for(self.loop.sock_recv(sck, bufsize), timeout)
        return await self.loop.sock_recv(sck, bufsize)

    @_async_sockwrapper_auto_connect()
    async def recv_into(self, buf: bytearray, sock: OpAnySocket = None, **kwargs) -> int:
        return await self.loop.sock_recv_into(self.socket if sock is None else sock, buf)

    @_async_sockwrapper_auto_connect()
    async def send(self, data: Union[str, bytes], flags: int = None, sock: OpAnySocket = None, timeout: Union[float, int] = AUTO, **kwargs):
        return await self.send_timeout(data, flags, sock, timeout, **kwargs)

    @_async_sockwrapper_auto_connect()
    async def sendall(
                self, data: Union[str, bytes], flags: int = None, sock: OpAnySocket = None, timeout: Union[float, int] = AUTO, **kwargs
            ):
        timeout, sck = self.send_timeout if timeout is AUTO else timeout, self.socket if sock is None else sock
        if timeout not in [None, False]:
            return await asyncio.wait_for(self.loop.sock_sendall(sck, byteify(data)), timeout)
        return await self.loop.sock_sendall(sck, byteify(data))

    @_async_sockwrapper_auto_connect()
    async def sendfile(
            self, file: IO[bytes], offset: int = None, count: int = None, fallback: bool = True, sock: OpAnySocket = None,
            timeout: Union[float, int] = AUTO, **kwargs
    ):
        timeout, sck = self.send_timeout if timeout is AUTO else timeout, self.socket if sock is None else sock
        if timeout not in [None, False]:
            return await asyncio.wait_for(self.loop.sock_sendfile(sck, file, offset=offset, count=count, fallback=fallback), timeout)
        return await self.loop.sock_sendfile(sck, file, offset=offset, count=count, fallback=fallback)

    @_async_sockwrapper_auto_connect()
    async def query(self, data: Union[str, bytes], bufsize: int = 32, eof_timeout=30, sock: OpAnySocket = None, **kwargs):
        timeout_fail, send_flags = kwargs.pop('timeout_fail', False), kwargs.pop('send_flags', kwargs.get('flags', None))
        recv_flags = kwargs.pop('recv_flags', kwargs.pop('flags', None))
        shared_timeout = kwargs.pop('timeout', AUTO)
        log.debug(" >> Sending %s bytes to %s:%s", len(data), self.host, self.port)
        snd_tmout, rcv_tmout = kwargs.pop('send_timeout', shared_timeout), kwargs.pop('read_timeout', shared_timeout)
        await self.sendall(byteify(data), flags=send_flags, sock=self.socket if sock is None else sock, timeout=snd_tmout)
        log.debug(" >> Reading %s bytes per chunk from %s:%s", bufsize, self.host, self.port)
        return await self.read_eof(
            bufsize, eof_timeout=eof_timeout, flags=recv_flags, timeout_fail=timeout_fail,
            sock=self.socket if sock is None else sock, read_timeout=rcv_tmout, **kwargs
        )

    @_async_sockwrapper_auto_connect()
    async def http_request(
            self, url="/", host=AUTO_DETECTED, method="GET", user_agent=settings.DEFAULT_USER_AGENT,
            extra_data: Union[STRBYTES, List[str]] = None, body: STRBYTES = None, eof_timeout=30, bufsize: int = 256,
            conv: Optional[Callable[[Union[bytes, str]], T]] = stringify, sock: OpAnySocket = None, **kwargs
    ) -> Union[str, bytes, T]:
        async with self:
            data = self._http_request(
                url, host=host, method=method, user_agent=user_agent, extra=extra_data, body=body, sock=sock, **kwargs
            )
            
            # await self.sendall(data)
            return await self.query(
                data, bufsize, eof_timeout=eof_timeout, timeout_fail=kwargs.get('timeout_fail', False), conv=conv, sock=sock, **kwargs
            )
            # return await super().http_request(
            #     url, host=host, method=method, user_agent=user_agent, extra=extra_data, body=body, eof_timeout=eof_timeout, **kwargs
            # )
    
    async def accept(self, sock: OpAnySocket = None, **kwargs) -> Tuple[AnySocket, Tuple[str, int]]:
        return await self.loop.sock_accept(self.socket if sock is None else sock)

    async def handle_connection(
            self, sock: AnySocket, addr: Tuple[str, int], callback: Callable[["AsyncSocketWrapper", Tuple[str, int]], Any],
            stop_return: Union[str, bytes] = None,
            **kwargs
    ):
        stop_compare, stop_compare_lower = kwargs.get('stop_compare', 'equal'), kwargs.get('stop_compare_lower', True)
        if stop_return is not None: stop_return = stringify(stop_return)
        log.info("[async] NEW CONNECTION: %s || %s", sock, addr)
        log.info("[async] Running callback: %s(%s, %s)\n", callback.__name__, sock, addr)
        orig_cres = await await_if_needed(callback(self.from_socket(sock), addr))
        cres = stringify(orig_cres)
        log.info("[async] Callback return data: %s\n\n", cres)
        if stop_return is not None:
            if stop_compare_lower: stop_return, cres = stop_return.lower(), cres.lower()
        
            if stop_compare.lower() in ['in', 'contain', 'contains', 'contained', 'within', 'inside']:
                if stop_return in cres or strip_null(stop_return) in strip_null(cres):
                    raise StopLoopOnMatch("[async] Matched stop_return. Parent should stop loop.", cres, stop_compare, stop_compare_lower)
        
            if cres == stop_return or strip_null(cres) == strip_null(stop_return):
                raise StopLoopOnMatch("[async] Matched stop_return. Parent should stop loop.", cres, stop_compare, stop_compare_lower)
        return orig_cres
    
    @_sockwrapper_auto_connect()
    async def on_connect(
            self, callback: Callable[["AsyncSocketWrapper", Tuple[str, int]], Any], timeout: AnyNum = None,
            stop_return: Union[str, bytes] = None, sock: OpAnySocket = None, **kwargs
    ):
        if not self.server:
            raise ValueError("This AsyncSocketWrapper has 'server' set to False. Can't handle incoming connections.")
        if not self.binded: self.bind(sock=self.socket if sock is None else sock)
        if not self.listening: self.listen(self.listen_backlog, sock=self.socket if sock is None else sock)
        # if stop_return is not None: stop_return = stringify(stop_return)
        # stop_compare, stop_compare_lower = kwargs.get('stop_compare', 'equal'), kwargs.get('stop_compare_lower', True)
        stop_return_match = None
        
        while self.connected and stop_return_match is None:
            log.info("[async] Waiting for incoming connection ( %s:%s || %s ) ...", self.host, self.port, self.socket.getsockname())
            sock, addr = await self.accept()
            try:
                with sock:
                    await self.handle_connection(sock, addr, callback, stop_return, **kwargs)
            except StopLoopOnMatch as e:
                log.info(" !!! Stopping on_connect as 'stop_return' has been matched: %s", stop_return)
                log.info(" !!! The matching message was: %s", e.match)
                break
            # log.info("[async] NEW CONNECTION: %s || %s", sock, addr)
            # log.info("[async] Running callback: %s(%s, %s)\n", callback.__name__, sock, addr)
            # cres = await await_if_needed(callback(self.from_socket(sock), addr))
            # cres = stringify(cres)
            # log.info("[async] Callback return data: %s\n\n", cres)
            # if stop_return is not None:
            #     if stop_compare_lower:
            #         stop_return, cres = stop_return.lower(), cres.lower()
            #     if stop_compare.lower() in ['in', 'contain', 'contains', 'contained', 'within', 'inside']:
            #         if stop_return in cres or strip_null(stop_return) in strip_null(cres):
            #             stop_return_match = cres
            #             break
            #     if cres == stop_return or strip_null(cres) == strip_null(stop_return):
            #         stop_return_match = cres
            #         break
        # if stop_return_match is not None:
        #     log.info(" !!! Stopping on_connect as 'stop_return' has been matched: %s", stop_return)
        #     log.info(" !!! The matching message was: %s", stop_return_match)
        
        log.info(" !!! Disconnected. Stopping on_connect.")

    async def __aenter__(self):
        # if not self._socket_ctx_mgr:
        #     self._socket_ctx_mgr = SocketContextManager(self)
        # if not self._layer_context:
        #     self._layer_context = LayeredContext(self._socket_ctx_mgr, max_layers=1)
        # return await self._layer_context.__aenter__()
        return await self.tracker.__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        # return await self._layer_context.__aexit__(exc_type, exc_val, exc_tb)
        return await self.tracker.__aexit__(exc_type, exc_val, exc_tb)


async def send_data_async(
        host: str, port: int, data: Union[bytes, str, Iterable], timeout: AnyNum = None, **kwargs
) -> Optional[Union[str, bytes]]:
    """
    
        >>> await send_data_async('termbin.com', 9999, "hello world\\nthis is a test\\n\\nlorem ipsum dolor\\n")
        'https://termbin.com/oi07'
    
    :param host:
    :param port:
    :param data:
    :param timeout:
    :param kwargs:
    :return:
    """
    fhost = f"({host}):{port}"
    chunk_size = int(kwargs.get('chunk', kwargs.get('chunk_size', 64)))
    string_result = is_true(kwargs.get('string_result', True))
    strip_result = is_true(kwargs.get('strip_result', True))
    fail = is_true(kwargs.get('fail', True))
    ip_version = kwargs.get('ip_version', 'any')
    timeout = empty_if(timeout, empty_if(socket.getdefaulttimeout(), 15, zero=True), zero=True)
    
    is_iter, data_iter = False, None
    
    if data is not None:
        if isinstance(data, (str, bytes, int, float)):
            data = byteify(data)
        else:
            try:
                data_iter = iter(data)
                is_iter = True
            except TypeError:
                # noinspection PyTypeChecker
                data = byteify(data)
        
    loop = asyncio.get_event_loop()
    try:
        s_ver = socket.AF_INET
        ip = await resolve_ip_async(host, ip_version)
        
        if ip_is_v6(ip): s_ver = socket.AF_INET6
        
        fhost += f" (IP: {ip})"
        
        with socket.socket(s_ver, socket.SOCK_STREAM) as s:
            s.settimeout(float(timeout))
            log.debug(" [...] Connecting to host: %s", fhost)
            await loop.sock_connect(s, (ip, port))
            log.debug(" [+++] Connected to %s\n", fhost)

            if data is None:
                log.debug(" [!!!] 'data' is None. Not transmitting any data to the host.")
            elif is_iter:
                i = 1
                for c in data_iter:
                    log.debug(" [...] Sending %s byte chunk (%s)\n", len(c), i)
                    await loop.sock_sendall(s, c)
            else:
                # We use 'sendall' to reliably send the entire contents of 'data' to the service we're connected to.
                log.debug(" [...] Sending %s bytes to %s ...\n", len(data), fhost)
                await loop.sock_sendall(s, data)
            # s.sendall(data)
            log.debug(" >> Reading response ...")
            res = b''
            i = 1
            while True:
                chunk = await loop.sock_recv(s, chunk_size)
                if not chunk: break
                res += chunk
                log.debug(" [...] Read %s byte chunk (%s)\n", len(chunk), i)
                i += 1
            if string_result:
                res = stringify(res)
                if strip_result: res = res.strip("\x00").strip().strip("\x00").strip()
            log.debug(" [+++] Got result ( %s bytes ) \n", len(res))
    except (socket.timeout, ConnectionRefusedError, ConnectionResetError, socket.gaierror) as e:
        if fail:
            raise e
        log.warning("Exception while connecting + sending data to: %s - reason: %s %s", fhost, type(e), str(e))
        return None
    return res


def send_data(
        host: str, port: int, data: Optional[Union[bytes, str, Iterable]] = None, timeout: Union[int, float] = None, **kwargs
) -> Optional[Union[str, bytes]]:
    """
        >>> from privex.helpers import send_data
        >>> send_data('termbin.com', 9999, "hello world\\nthis is a test\\n\\nlorem ipsum dolor\\n")
        'https://termbin.com/oi07'
    
    :param str host: The hostname or IPv4/v6 address to connect to
    :param port:     The port number to connect to on ``host``
    :param bytes|str|iter data:      The data to send to ``host:port`` via a TCP socket. Generally :class:`bytes` / :class:`str`.
                                     Can be an iterator/generator to send data in chunks. Can be ``None`` to disable sending data, instead
                                     only receiving and returning data.
    :param float|int timeout: Socket timeout. If not passed, uses the default from :func:`socket.getdefaulttimeout`.
                              If the global default timeout is ``None``, then falls back to ``15``
    :param kwargs:
    :keyword int chunk: (Default: ``64``) Maximum number of bytes to read into buffer per socket receive call.
    :keyword bool string_result: (Default: ``True``) If ``True``, the response sent by the server will be casted into a :class:`str`
                                 before returning it.
    :keyword bool strip_result: (Default: ``True``) This argument only works if ``string_result`` is also True.
                                If both ``string_result`` and ``strip_result`` are ``True``, the response sent by the server will
                                have whitespace, newlines, and null bytes trimmed from the start and end after it's casted into a string.
    :keyword bool fail: (Default: ``True``) If ``True``, will raise exceptions when connection errors occur. When ``False``, will simply
                        ``None`` if there are connection exceptions raised during this function's execution.
    :keyword str|int ip_version: (Default: ``any``)
    :return:
    """
    fhost = f"({host}):{port}"
    chunk_size = int(kwargs.get('chunk', kwargs.get('chunk_size', 64)))
    string_result = is_true(kwargs.get('string_result', True))
    strip_result = is_true(kwargs.get('strip_result', True))
    fail = is_true(kwargs.get('fail', True))
    ip_version = kwargs.get('ip_version', 'any')
    timeout = empty_if(timeout, empty_if(socket.getdefaulttimeout(), 15, zero=True), zero=True)

    is_iter, data_iter, is_v6, v4_address, host_is_ip = False, None, False, None, False

    if data is not None:
        if isinstance(data, (str, bytes, int, float)):
            data = byteify(data)
        else:
            try:
                data_iter = iter(data)
                is_iter = True
            except TypeError:
                # noinspection PyTypeChecker
                data = byteify(data)
    
    try:
        ip_network(host)
        host_is_ip = True
    except (TypeError, ValueError) as e:
        host_is_ip = False
    
    try:
        # First we resolve the IP address of 'host', so we can detect whether we're connecting to an IPv4 or IPv6 host,
        # letting us adjust the AF_INET variable accordingly.
        s_ver = socket.AF_INET
        ip = resolve_ip(host, ip_version)
    
        if ip_is_v6(ip):
            s_ver, is_v6 = socket.AF_INET6, True
            if not host_is_ip:
                try:
                    v4_address = resolve_ip(host, 'v4')
                except (socket.timeout, ConnectionRefusedError, ConnectionResetError, socket.gaierror, AttributeError) as e:
                    log.warning(
                        "Warning: failed to resolve IPv4 address for %s (to be used as a backup if IPv6 is broken). Reason: %s %s ",
                        type(e), str(e)
                    )
        
        fhost += f" (IP: {ip})"

    except (socket.timeout, ConnectionRefusedError, ConnectionResetError, socket.gaierror) as e:
        if fail:
            raise e
        log.warning("Exception while connecting + sending data to: %s - reason: %s %s", fhost, type(e), str(e))
        return None
    
    try:
        with socket.socket(s_ver, socket.SOCK_STREAM) as s:
            # Once we have our socket object, we set the timeout (by default it could hang forever), and open the connection.
            s.settimeout(timeout)
            log.debug(" [...] Connecting to host: %s", fhost)
            s.connect((ip, port))
            log.debug(" [+++] Connected to %s\n", fhost)
            
            if data is None:
                log.debug(" [!!!] 'data' is None. Not transmitting any data to the host.")
            elif is_iter:
                i = 1
                for c in data_iter:
                    log.debug(" [...] Sending %s byte chunk (%s)\n", len(c), i)
                    s.sock_sendall(c)
            else:
                # We use 'sendall' to reliably send the entire contents of 'data' to the service we're connected to.
                log.debug(" [...] Sending %s bytes to %s ...\n", len(data), fhost)
                s.sendall(data)
            # Once we've sent 'data',
            log.debug(" >> Reading response ...")
            res = b''
            i = 1
            while True:
                chunk = s.recv(chunk_size)
                if not chunk: break
                res += chunk
                log.debug(" [...] Read %s byte chunk (%s)\n", len(chunk), i)
                i += 1
            if string_result:
                res = stringify(res)
                if strip_result: res = res.strip("\x00").strip().strip("\x00").strip()
            log.debug(" [+++] Got result ( %s bytes ) \n", len(res))
    except (socket.timeout, ConnectionRefusedError, ConnectionResetError, socket.gaierror) as e:
        log.warning("Exception while connecting + sending data to: %s - reason: %s %s", fhost, type(e), str(e))
        if is_v6 and not empty(v4_address):
            log.warning(
                "Retrying connection to %s over IPv4 instead of IPv6. || IPv6 address: %s || IPv4 address: %s ",
                fhost, ip, v4_address
            )
            return send_data(host, port, data, timeout=timeout, **kwargs)

        if fail:
            raise e
        return None
    return res


def upload_termbin(data: Union[bytes, str], timeout: Union[int, float] = None, **kwargs) -> str:
    """
    Upload the :class:`bytes` / :class:`string` ``data`` to the pastebin service `TermBin`_ ,
    using the hostname and port defined in :attr:`privex.helpers.settings.TERMBIN_HOST`
    and :attr:`privex.helpers.settings.TERMBIN_PORT`
    
    NOTE - An AsyncIO version of this function is available: :func:`.upload_termbin_async`
    
    Returns the `TermBin`_ URL as a string - which is a raw download / viewing link for the paste.
    
    .. _TermBin:   https://termbin.com
    
        >>> my_data = "hello world\\nthis is a test\\n\\nlorem ipsum dolor\\n"
        >>> upload_termbin(my_data)
        'https://termbin.com/kerjk'
    
    :param bytes|str data:    The data to upload to `TermBin`_ - as either :class:`str` or :class:`bytes`
    :param float|int timeout: Socket timeout. If not passed, uses the default from :func:`socket.getdefaulttimeout`.
                              If the global default timeout is ``None``, then falls back to ``15``
    :return str url:   The `TermBin`_ URL to your paste as a string - which is a raw download / viewing link for the paste.
    """
    data = byteify(data)
    log.info(" [...] Uploading %s bytes to termbin ...\n", len(data))
    res = send_data(settings.TERMBIN_HOST, settings.TERMBIN_PORT, data, timeout=timeout, **kwargs)
    log.info(" [+++] Got termbin link: %s \n", res)
    
    return res


def upload_termbin_file(filename: str, timeout: int = 15, **kwargs) -> str:
    """
    Uploads the file ``filename`` to `TermBin`_ and returns the paste URL as a string.

    .. NOTE:: An AsyncIO version of this function is available: :func:`.upload_termbin_file_async`

    .. NOTE:: If the data you want to upload is already loaded into a variable - you can use :func:`.upload_termbin` instead,
              which accepts your data directly - through a :class:`str` or :class:`bytes` parameter

    .. _TermBin:   https://termbin.com

    :param str filename:      The path (absolute or relative) to the file you want to upload to `TermBin`_ - as a :class:`str`
    :param float|int timeout: Socket timeout. If not passed, uses the default from :func:`socket.getdefaulttimeout`.
                              If the global default timeout is ``None``, then falls back to ``15``
    :return str url:   The `TermBin`_ URL to your paste as a string - which is a raw download / viewing link for the paste.
    """
    log.info(" >> Uploading file '%s' to termbin", filename)
    
    with open(filename, 'rb') as fh:
        log.debug(" [...] Opened file %s - reading contents into RAM...", filename)
        data = fh.read()
        log.debug(" [+++] Loaded file into RAM. Total size: %s bytes", len(data))
    
    res = upload_termbin(data, timeout=timeout, **kwargs)
    log.info(" [+++] Uploaded file %s to termbin. Got termbin link: %s \n", filename, res)
    return res


async def upload_termbin_async(data: Union[bytes, str], timeout: Union[int, float] = None) -> str:
    """
    Upload the :class:`bytes` / :class:`string` ``data`` to the pastebin service `TermBin`_ ,
    using the hostname and port defined in :attr:`privex.helpers.settings.TERMBIN_HOST`
    and :attr:`privex.helpers.settings.TERMBIN_PORT`

    NOTE - A synchronous (non-async) version of this function is available: :func:`.upload_termbin`

    Returns the `TermBin`_ URL as a string - which is a raw download / viewing link for the paste.

    .. _TermBin:   https://termbin.com

        >>> my_data = "hello world\\nthis is a test\\n\\nlorem ipsum dolor\\n"
        >>> await upload_termbin_async(my_data)
        'https://termbin.com/kerjk'

    :param bytes|str data:    The data to upload to `TermBin`_ - as either :class:`str` or :class:`bytes`
    :param float|int timeout: Socket timeout. If not passed, uses the default from :func:`socket.getdefaulttimeout`.
                              If the global default timeout is ``None``, then falls back to ``15``
    :return str url:   The `TermBin`_ URL to your paste as a string - which is a raw download / viewing link for the paste.
    """
    data = byteify(data)
    log.info(" [...] Uploading %s bytes to termbin ...\n", len(data))
    res = await send_data_async(settings.TERMBIN_HOST, settings.TERMBIN_PORT, data, timeout=timeout)
    log.info(" [+++] Got termbin link: %s \n", res)
    
    return res


async def upload_termbin_file_async(filename: str, timeout: int = 15) -> str:
    """
    Uploads the file ``filename`` to `TermBin`_ and returns the paste URL as a string.
    
    .. NOTE:: A synchronous (non-async) version of this function is available: :func:`.upload_termbin_file`
    
    .. NOTE:: If the data you want to upload is already loaded into a variable - you can use :func:`.upload_termbin_async` instead,
              which accepts your data directly - through a :class:`str` or :class:`bytes` parameter

    
    .. _TermBin:   https://termbin.com

    :param str filename:      The path (absolute or relative) to the file you want to upload to `TermBin`_ - as a :class:`str`
    :param float|int timeout: Socket timeout. If not passed, uses the default from :func:`socket.getdefaulttimeout`.
                              If the global default timeout is ``None``, then falls back to ``15``
    :return str url:   The `TermBin`_ URL to your paste as a string - which is a raw download / viewing link for the paste.
    """
    log.info(" >> Uploading file '%s' to termbin", filename)
    
    with open(filename, 'rb') as fh:
        log.debug(" [...] Opened file %s - reading contents into RAM...", filename)
        data = fh.read()
        log.debug(" [+++] Loaded file into RAM. Total size: %s bytes", len(data))
    
    res = await upload_termbin_async(data, timeout=timeout)
    log.info(" [+++] Uploaded file %s to termbin. Got termbin link: %s \n", filename, res)
    return res



import asyncio
import pickle
from typing import Any, Union, Optional

from aioredis.commands import ContextRedis
from async_property import async_property

from privex.helpers.common import empty

# from privex.helpers import plugin
# from privex.helpers.cache.CacheAdapter import CacheAdapter
from privex.helpers.exceptions import CacheNotFound
from privex.helpers.settings import DEFAULT_CACHE_TIMEOUT
from privex.helpers.types import VAL_FUNC_CORO

# if plugin.HAS_ASYNC_REDIS:
from privex.helpers.plugin import get_redis_async, close_redis_async
from privex.helpers.cache.asyncx.base import AsyncCacheAdapter
import aioredis
import logging

log = logging.getLogger(__name__)


class AsyncRedisCache(AsyncCacheAdapter):
    """
    A Redis backed implementation of :class:`.AsyncCacheAdapter`. Uses the global Redis instance from
    :py:mod:`privex.helpers.plugin` by default, however custom Redis instances can be passed in via
    the constructor argument ``redis_instance``.

    To allow for a wide variety of Python objects to be safely stored and retrieved from Redis, this class
    uses the :py:mod:`pickle` module for serialising + un-serialising values to/from Redis.

    **Basic Usage**::

        >>> from privex.helpers import AsyncRedisCache
        >>> rc = AsyncRedisCache()
        >>> await rc.set('hello', 'world')
        >>> rc['hello']
        'world'


    **Disabling Pickling**

    In some cases, you may need interoperable caching with other languages. The :py:mod:`pickle` serialisation
    technique is extremely specific to Python and is largely unsupported outside of Python. Thus if you need
    to share Redis cache data with applications in other languages, then you must disable pickling.

    **WARNING:** If you disable pickling, then you must perform your own serialisation + de-serialization on
    complex objects such as ``dict``, ``list``, ``Decimal``, or arbitrary classes/functions after getting
    or setting cache keys.

    **Disabling Pickle per instance**

    Pass ``use_pickle=False`` to the constructor, or access the attribute directly to disable pickling for a
    single instance of RedisCache (not globally)::

        >>> rc = AsyncRedisCache(use_pickle=False)  # Opt 1. Disable pickle in constructor
        >>> rc.use_pickle = False                   # Opt 2. Disable pickle on an existing instance


    **Disabling Pickle by default on any new instances**

    Change the static attribute :py:attr:`.pickle_default` to ``False`` to disable the use of pickle by default
    across any new instances of RedisCache::

        >>> AsyncRedisCache.pickle_default = False

    """
    
    pickle_default: bool = True
    """
    Change this to ``False`` to disable the use of :py:mod:`pickle` by default for any new instances
    of this class.
    """
    
    use_pickle: bool
    """If ``True``, will use :py:mod:`pickle` for serializing objects before inserting into Redis, and
    un-serialising objects retrieved from Redis. This attribute is set in :py:meth:`.__init__`.

    Change this to ``False`` to disable the use of :py:mod:`pickle` - instead values will be passed to / returned
    from Redis as-is, with no serialisation (this may require you to manually serialize complex types such
    as ``dict`` and ``Decimal`` before insertion, and un-serialise after retrieval).
    """
    adapter_enter_reconnect: bool = True
    adapter_exit_close: bool = True
    
    _redis_conn: Optional[Union[aioredis.Redis, aioredis.ConnectionsPool]]
    _redis: Optional[ContextRedis]
    
    def __init__(self, use_pickle: bool = None, redis_instance: aioredis.Redis = None, *args, **kwargs):
        """
        RedisCache by default uses the global Redis instance from :py:mod:`privex.helpers.plugin`.

        It's recommended to use :py:func:`privex.helpers.plugin.configure_redis` if you need to change any
        Redis settings, as this will adjust the global settings and re-instantiate the global instance if required.

        Alternatively, you may pass an instance of :class:`redis.Redis` as ``redis_instance``, then that will
        be used instead of the global instance from :py:func:`.get_redis`

        :param bool use_pickle: (Default: ``True``) Use the built-in ``pickle`` to serialise values before
                                storing in Redis, and un-serialise when loading from Redis

        :param redis.Redis redis_instance: If this isn't ``None`` / ``False``, then this Redis instance will be
                                           used instead of the global one from :py:func:`.get_redis`
        
        :keyword bool enter_reconnect: Pass ``enter_reconnect=False`` to disable calling :meth:`.reconnect` when entering this cache
                                       adapter as a context manager (:meth:`.__aenter__`)
        :keyword bool exit_close: Pass ``exit_close=False`` to disable calling :meth:`.close` when exiting this cache
                                  adapter as a context manager (:meth:`.__aexit__`)
        """
        super().__init__(*args, **kwargs)
        self._redis = None if not redis_instance else redis_instance
        self._redis_conn = None
        self.use_pickle = self.pickle_default if use_pickle is None else use_pickle
    
    @async_property
    async def redis(self) -> aioredis.Redis:
        # if self._redis_conn is None:
        #     self._redis_conn = await get_redis_async()
        # if self._redis is not None:
        #     return self._redis
        # return self._redis
        # if self._redis is None:
        #     self._redis = await self.connect()
        return await self.connect()

    async def get(self, key: str, default: Any = None, fail: bool = False) -> Any:
        key = str(key)
        r = await self.redis
        res = await r.get(key)
        if empty(res):
            if fail: raise CacheNotFound(f'Cache key "{key}" was not found.')
            return default
        return pickle.loads(res) if self.use_pickle else res
    
    async def set(self, key: str, value: Any, timeout: Optional[int] = DEFAULT_CACHE_TIMEOUT):
        r: aioredis.Redis = await self.redis
        v = pickle.dumps(value) if self.use_pickle else value
        return await r.set(str(key), v, expire=timeout)

    # async def get_or_set(self, key: str, value: VAL_FUNC_CORO, timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
    #     return await self.get_or_set_async(key=key, value=value, timeout=timeout)
    
    async def remove(self, *key: str) -> bool:
        removed = 0
        for k in key:
            k = str(k)
            try:
                await self.get(k, fail=True)
                r: aioredis.Redis = await self.redis
                await r.delete(k)
                removed += 1
            except CacheNotFound:
                pass
        
        return removed == len(key)
    
    async def update_timeout(self, key: str, timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
        key, timeout = str(key), int(timeout)
        v = await self.get(key=key, fail=True)
        await self.set(key=key, value=v, timeout=timeout)
        return v
    
    async def connect(self, *args, **kwargs) -> ContextRedis:
        if not self._redis_conn:
            self._redis_conn = await get_redis_async()
        if not self._redis:
            self._redis = await self._redis_conn
        return self._redis
    
    async def close(self):
        if self._redis is not None:
            log.debug("Closing AsyncIO Redis instance %s._redis", self.__class__.__name__)
            self._redis.close()
            self._redis = None
        # Closing the Redis connection directly from this method usually leads to problems...
        # It's safest to just set it to None and then call close_redis_async()
        if self._redis_conn is not None:
            log.debug("Clearing AsyncIO Redis connection pool %s._redis_conn", self.__class__.__name__)
            self._redis_conn = None
        await close_redis_async()




import asyncio
from abc import ABC, abstractmethod
from typing import Any, Optional, Union, Coroutine, Awaitable

from privex.helpers.asyncx import await_if_needed
from privex.helpers.common import empty_if

from privex.helpers.exceptions import CacheNotFound
from privex.helpers.settings import DEFAULT_CACHE_TIMEOUT
from privex.helpers.types import VAL_FUNC_CORO


class CacheAdapter(ABC):
    """
    **CacheAdapter** is an abstract base class which scaffolds methods for implementing a Cache, allowing for
    consistent methods and method signatures across all child classes which implement it.
    
    This class cannot be instantiated by itself, only child classes which extend :class:`.CacheAdapter` and implement
    all methods marked with ``@abstractmethod`` can be instantiated.
    
    For an example implementation of CacheAdapter, see :class:`privex.helpers.cache.MemoryCache`
    
    """

    adapter_enter_reconnect: bool = True
    """
    Controls whether :meth:`.__enter__` automatically calls :meth:`.reconnect` to clear and re-create any previous
    connections/instances for the adapter.
    """
    adapter_exit_close: bool = True
    """
    Controls whether :meth:`.__exit__` automatically calls :meth:`.close` to close any connections/instances and destroy
    library class instances from the current adapter instance.
    """

    ins_enter_reconnect: bool
    """
    Per-instance version of :attr:`.adapter_enter_reconnect`, which is set via ``enter_reconnect`` the constructor.
    When ``__init__`` ``enter_reconnect`` is empty, it inherits the class attribute  value from :attr:`.adapter_enter_reconnect`
    """
    ins_exit_close: bool
    """
    Per-instance version of :attr:`.adapter_exit_close`, which is set via ``exit_close`` the constructor.
    When ``__init__`` ``exit_close`` is empty, it inherits the class attribute  value from :attr:`.adapter_exit_close`
    """
    
    def __init__(self, *args, enter_reconnect: Optional[bool] = None, exit_close: Optional[bool] = None, **kwargs):
        self.ins_enter_reconnect = empty_if(enter_reconnect, self.adapter_enter_reconnect)
        self.ins_exit_close = empty_if(exit_close, self.adapter_exit_close)
    
    @abstractmethod
    def get(self, key: str, default: Any = None, fail: bool = False) -> Any:
        """
        Return the value of cache key ``key``. If the key wasn't found, or it was expired, then ``default`` will be
        returned.
        
        Optionally, you may choose to pass ``fail=True``, which will cause this method to raise :class:`.CacheNotFound`
        instead of returning ``default`` when a key is non-existent / expired.
        
        :param str key: The cache key (as a string) to get the value for, e.g. ``example:test``
        :param Any default: If the cache key ``key`` isn't found / is expired, return this value (Default: ``None``)
        :param bool fail: If set to ``True``, will raise :class:`.CacheNotFound` instead of returning ``default``
                          when a key is non-existent / expired.
        
        :raises CacheNotFound: Raised when ``fail=True`` and ``key`` was not found in cache / expired.
        
        :return Any value: The value of the cache key ``key``, or ``default`` if it wasn't found.
        """
        raise NotImplemented(f'{self.__class__.__name__} must implement .get()')

    @abstractmethod
    def set(self, key: str, value: Any, timeout: Optional[int] = DEFAULT_CACHE_TIMEOUT):
        """
        Set the cache key ``key`` to the value ``value``, and automatically expire the key after ``timeout`` seconds
        from now.
        
        If ``timeout`` is ``None``, then the key will never expire (unless the cache implementation loses it's
        persistence, e.g. memory caches with no disk writes).
        
        :param str key: The cache key (as a string) to set the value for, e.g. ``example:test``
        :param Any value: The value to store in the cache key ``key``
        :param int timeout: The amount of seconds to keep the data in cache. Pass ``None`` to disable expiration.
        """
        raise NotImplemented(f'{self.__class__.__name__} must implement .set()')

    @abstractmethod
    def remove(self, *key: str) -> bool:
        """
        Remove one or more keys from the cache.
        
        If all cache keys existed before removal, ``True`` will be returned. If some didn't exist (and thus couldn't
        remove), then ``False`` will be returned.
        
        :param str key: The cache key(s) to remove
        :return bool removed: ``True`` if ``key`` existed and was removed
        :return bool removed: ``False`` if ``key`` didn't exist, and no action was taken.
        """
        raise NotImplemented(f'{self.__class__.__name__} must implement .remove()')

    @abstractmethod
    def update_timeout(self, key: str, timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
        """
        Update the timeout for a given ``key`` to ``datetime.utcnow() + timedelta(seconds=timeout)``
        
        This method should accept keys which are already expired, allowing expired cache keys to have their timeout
        extended **after** expiry.
        
        **Example**::
        
            >>> c = CacheAdapter()
            >>> c.set('example', 'test', timeout=60)
            >>> sleep(70)
            >>> c.update_timeout('example', timeout=60)   # Reset the timeout for ``'example'`` to ``now + 60 seconds``
            >>> c.get('example')
            'test'
        
        :param str key: The cache key to update the timeout for
        :param int timeout: Reset the timeout to this many seconds from ``datetime.utcnow()``
        :raises CacheNotFound: Raised when ``key`` was not found in cache (thus cannot extend timeout)
        :return Any value: The value of the cache key
        """
        raise NotImplemented(f'{self.__class__.__name__} must implement .extend_timeout()')

    def get_or_set(self, key: str, value: Union[Any, callable], timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
        """
        Attempt to return the value of ``key`` in the cache. If ``key`` doesn't exist or is expired, then it will be
        set to ``value``, and ``value`` will be returned.

        The ``value`` parameter can be any standard type such as ``str`` or ``dict`` - or it can be a callable
        function / method which returns the value to set and return.

        **Basic Usage**::

            >>> c = CacheAdapter()
            >>> c.get('testing')
            None
            >>> c.get_or_set('testing', 'hello world')
            'hello world'
            >>> c.get('testing')
            'hello world'

        **Set and get the value from a function if ``key`` didn't exist / was expired**::

            >>> def my_func(key): return "hello {} world".format(key)
            >>> c = CacheAdapter()
            >>> c.get_or_set('example', my_func)
            'hello example world'
            >>> c.get('example')
            'hello example world'

        :param str key: The cache key (as a string) to get/set the value for, e.g. ``example:test``
        :param Any value: The value to store in the cache key ``key``. Can be a standard type, or a callable function.
        :param int timeout: The amount of seconds to keep the data in cache. Pass ``None`` to disable expiration.
        :return Any value: The value of the cache key ``key``, or ``value`` if it wasn't found.
        """
        key, timeout = str(key), int(timeout)
        try:
            k = self.get(key, fail=True)
        except CacheNotFound:
            k = value(key) if callable(value) else value
            self.set(key=key, value=k, timeout=timeout)
        return k

    async def get_or_set_async(self, key: str, value: VAL_FUNC_CORO, timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
        """
        Async coroutine compatible version of :meth:`.get_or_set`.
        
        **Example with Async function**::
        
            >>> async def my_coro(key): return f"hello {key} world"
            >>> c = CacheAdapter()
            >>> await c.get_or_set_async('coro_example', my_coro)
            'hello example world'
            >>> c.get('coro_example')
            'hello example world'
        
        **Also works with non-async functions**::
            
            >>> def my_func(key): return f"hello {key} world"
            >>> await c.get_or_set_async('func_example', my_func)
            'hello example world'
            >>> c.get('func_example')
            'hello example world'
        
        :param str key: The cache key (as a string) to get/set the value for, e.g. ``example:test``
        :param Any value: The value to store in the cache key ``key``. Can be a standard type, a coroutine / awaitable,
                          or a plain callable function.
        :param int timeout: The amount of seconds to keep the data in cache. Pass ``None`` to disable expiration.
        :return Any value: The value of the cache key ``key``, or ``value`` if it wasn't found.
        """
        key, timeout = str(key), int(timeout)
        try:
            k = self.get(key, fail=True)
        except CacheNotFound:
            k = value
            if asyncio.iscoroutinefunction(value):
                k = await value(key)
            if asyncio.iscoroutine(value):
                k = await value
            elif callable(value):
                k = value(key)
            self.set(key=key, value=k, timeout=timeout)
        return k

    def close(self, *args, **kwargs) -> Any:
        """
        Close any cache library connections, and destroy their local class instances by setting them to ``None``.
        """
        return f"close() is not implemented by {self.__class__.__name__}"

    def connect(self, *args, **kwargs) -> Any:
        """
        Create an instance of the library used to interact with the caching system, ensure it's connection is open,
        and store the instance on this class instance - only if not already connected.

        Should return the class instance which was created.
        """
        return f"close() is not implemented by {self.__class__.__name__}"

    def reconnect(self, *args, **kwargs) -> Any:
        """
        Calls :meth:`.close` to close any previous connections and cleanup instances, then re-create the
        connection(s)/instance(s) by calling :meth:`.connect`
        """
        self.close()
        return self.connect()

    def __getitem__(self, item):
        try:
            return self.get(key=item, fail=True)
        except CacheNotFound:
            raise KeyError(f'Key "{item}" not found in cache.')

    def __setitem__(self, key, value):
        return self.set(key=key, value=value)
    
    async def __aenter__(self):
        """
        Before starting a context manager, we close and cleanup any previous connection and re-create a fresh connection
        and instance, ensuring no conflicts such as connections/instances attached to other AsyncIO event loops :)
        """
        if self.ins_enter_reconnect:
            await await_if_needed(self.reconnect())
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Once a context manager is finished, we close and cleanup any instances / connections, ensuring no conflicts
        for code that might want to use the cache adapter in a different event loops.
        """
        if self.ins_exit_close:
            await await_if_needed(self.reconnect())
        return None
    
    def __enter__(self):
        """
        Before starting a context manager, we close and cleanup any previous connection and re-create a fresh connection
        and instance, ensuring no conflicts such as connections/instances attached to other threads :)
        """
        if self.ins_enter_reconnect:
            self.reconnect()
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Once a context manager is finished, we close and cleanup any instances / connections, ensuring no conflicts
        for code that might want to use the cache adapter in a different threads / event loops.
        """
        if self.ins_exit_close:
            self.close()
        return None




from decimal import Decimal
from ipaddress import IPv4Address, IPv6Address, IPv4Network, IPv6Network
from typing import Union, Coroutine, Awaitable, Any, TypeVar, Callable

Number = Union[Decimal, int, float]
"""Shorter alias for ``Union[Decimal, int, float]``"""

AnyNum = NumberStr = Union[Decimal, int, float, str]
"""Shorter alias for ``Union[Decimal, int, float, str]``"""

VAL_FUNC_CORO = Union[Any, callable, Coroutine, Awaitable]
"""Type alias for callable's, coroutine's and awaitable's"""

IP_OR_STR = Union[str, IPv4Address, IPv6Address]
"""Shorthand union which accepts :class:`.IPv4Address`, :class:`.IPv6Address` or :class:`str`"""
NET_OR_STR = Union[str, IPv4Network, IPv6Network]
"""Shorthand union which accepts :class:`.IPv4Network`, :class:`.IPv6Network` or :class:`str`"""
IP_NET_OR_STR = Union[str, IPv4Address, IPv6Address, IPv4Network, IPv6Network]
"""Combined :attr:`.IP_OR_STR` + :attr:`.NET_OR_STR`"""

T = TypeVar('T')
"""Plain generic type variable for use in helper functions"""
K = TypeVar('K')
"""Plain generic type variable for use in helper functions"""
V = TypeVar('V')
"""Plain generic type variable for use in helper functions"""
C = TypeVar('C', type, Callable)
"""Generic type variable constrained to :class:`type` / :class:`typing.Callable` for use in helper functions"""
CL = TypeVar('CL', type, Callable)
"""Generic type variable constrained to :class:`type` / :class:`typing.Callable` for use in helper functions"""


USE_ORIG_VAR = type('UseOrigVar', (), {})
"""
A simple functionless type, used purely as a default parameter value meaning "fallback to the value from a certain
other parameter".

Primarily used in :func:`.empty_if` but can be used by any function/method, including use outside of privex-helpers.
"""

NO_RESULT = type('NoResult', (), {})
"""
Simple functionless type which means "no results were found or nothing matched this function's query".

Useful for returning a unique "nothing to return" value from functions where ``None`` / ``False`` might be considered as successful,
and exceptions aren't suitable::

    >>> from privex.helpers.types import NO_RESULT
    >>> def some_func(x: int):
    ...     if (x + 1) > 2: return True
    ...     elif (x + 1) < 2: return False
    ...     if x == 0 or (x + 1) == 0: return None
    ...     return NO_RESULT
    >>> res = some_func(-2)
    >>> res == NO_RESULT
    True

"""

STRBYTES = Union[bytes, str]
"""Shorter alias for ``Union[bytes, str]``"""

AUTO = AUTOMATIC = AUTO_DETECTED = type('AutoDetected', (), {})
"""
Another functionless type, intended to stand-in as the default value for a parameter, with the
meaning "automatically populate this parameter from another source" e.g. instance state attributes
"""




import asyncio
import pickle
from typing import Any, Union, Optional
from async_property import async_property
from privex.helpers.common import empty, byteify
from privex.helpers.exceptions import CacheNotFound
from privex.helpers.settings import DEFAULT_CACHE_TIMEOUT
from privex.helpers.plugin import get_memcached_async, close_memcached_async
from privex.helpers.cache.asyncx.base import AsyncCacheAdapter
import aiomcache
import logging

log = logging.getLogger(__name__)


class AsyncMemcachedCache(AsyncCacheAdapter):
    """
    A Memcached backed implementation of :class:`.AsyncCacheAdapter`. Uses the global Memcached instance from
    :py:mod:`privex.helpers.plugin` by default, however custom Memcached instances can be passed in via
    the constructor argument ``mcache_instance``.

    To allow for a wide variety of Python objects to be safely stored and retrieved from Memcached, this class
    uses the :py:mod:`pickle` module for serialising + un-serialising values to/from Memcached.

    **Basic Usage**::

        >>> from privex.helpers import AsyncMemcachedCache
        >>> rc = AsyncMemcachedCache()
        >>> await rc.set('hello', 'world')
        >>> rc['hello']
        'world'


    **Disabling Pickling**

    In some cases, you may need interoperable caching with other languages. The :py:mod:`pickle` serialisation
    technique is extremely specific to Python and is largely unsupported outside of Python. Thus if you need
    to share Memcached cache data with applications in other languages, then you must disable pickling.

    **WARNING:** If you disable pickling, then you must perform your own serialisation + de-serialization on
    complex objects such as ``dict``, ``list``, ``Decimal``, or arbitrary classes/functions after getting
    or setting cache keys.

    **Disabling Pickle per instance**

    Pass ``use_pickle=False`` to the constructor, or access the attribute directly to disable pickling for a
    single instance of MemcachedCache (not globally)::

        >>> rc = AsyncMemcachedCache(use_pickle=False)  # Opt 1. Disable pickle in constructor
        >>> rc.use_pickle = False                       # Opt 2. Disable pickle on an existing instance


    **Disabling Pickle by default on any new instances**

    Change the static attribute :py:attr:`.pickle_default` to ``False`` to disable the use of pickle by default
    across any new instances of AsyncMemcachedCache::

        >>> AsyncMemcachedCache.pickle_default = False



    """
    
    pickle_default: bool = True
    """
    Change this to ``False`` to disable the use of :py:mod:`pickle` by default for any new instances
    of this class.
    """
    
    use_pickle: bool
    """If ``True``, will use :py:mod:`pickle` for serializing objects before inserting into Memcached, and
    un-serialising objects retrieved from Memcached. This attribute is set in :py:meth:`.__init__`.

    Change this to ``False`` to disable the use of :py:mod:`pickle` - instead values will be passed to / returned
    from Memcached as-is, with no serialisation (this may require you to manually serialize complex types such
    as ``dict`` and ``Decimal`` before insertion, and un-serialise after retrieval).
    """

    adapter_enter_reconnect: bool = True
    adapter_exit_close: bool = True

    _mcache: Optional[aiomcache.Client]
    
    def __init__(self, use_pickle: bool = None, mcache_instance: aiomcache.Client = None, *args, **kwargs):
        """
        AsyncMemcachedCache by default uses the global Memcached instance from :py:mod:`privex.helpers.plugin`.

        It's recommended to use :py:func:`privex.helpers.plugin.configure_memcached_async` if you need to change any
        Memcached settings, as this will adjust the global settings and re-instantiate the global instance if required.

        Alternatively, you may pass an instance of :class:`aiomcache.Client` as ``mcache_instance``, then that will
        be used instead of the global instance from :py:func:`.get_memcached_async`

        :param bool use_pickle: (Default: ``True``) Use the built-in ``pickle`` to serialise values before
                                storing in Memcached, and un-serialise when loading from Memcached

        :param aiomcache.Client mcache_instance: If this isn't ``None`` / ``False``, then this Memcached instance will be
                                                 used instead of the global one from :py:func:`.get_memcached_async`
        
        :keyword bool enter_reconnect: Pass ``enter_reconnect=False`` to disable calling :meth:`.reconnect` when entering this cache
                                       adapter as a context manager (:meth:`.__aenter__`)
        :keyword bool exit_close: Pass ``exit_close=False`` to disable calling :meth:`.close` when exiting this cache
                                  adapter as a context manager (:meth:`.__aexit__`)
        """
        super().__init__(*args, **kwargs)
        self._mcache = None if not mcache_instance else mcache_instance
        self.use_pickle = self.pickle_default if use_pickle is None else use_pickle
    
    @async_property
    async def mcache(self) -> aiomcache.Client:
        # We can't recycle connections with aiomcache, so we need to make sure to get a new connection for every instance.
        # if self._mcache is None:
        #     self._mcache = await get_memcached_async(new_connection=True)
        return await self.connect()
    
    async def get(self, key: Union[bytes, str], default: Any = None, fail: bool = False) -> Any:
        key = byteify(key)
        r = await self.mcache
        res = await r.get(key)
        if empty(res):
            if fail: raise CacheNotFound(f'Cache key "{key}" was not found.')
            return default
        return pickle.loads(res) if self.use_pickle else res
    
    async def set(self, key: Union[bytes, str], value: Any, timeout: Optional[int] = DEFAULT_CACHE_TIMEOUT):
        r: aiomcache.Client = await self.mcache
        v = pickle.dumps(value) if self.use_pickle else value
        return await r.set(byteify(key), v, exptime=timeout)
    
    async def remove(self, *key: Union[bytes, str]) -> bool:
        removed = 0
        for k in key:
            k = byteify(k)
            try:
                await self.get(k, fail=True)
                r = await self.mcache
                await r.delete(k)
                removed += 1
            except CacheNotFound:
                pass
        
        return removed == len(key)
    
    async def update_timeout(self, key: str, timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
        key, timeout = str(key), int(timeout)
        v = await self.get(key=key, fail=True)
        await self.set(key=key, value=v, timeout=timeout)
        return v

    async def connect(self, *args, new_connection=True, **kwargs) -> aiomcache.Client:
        # We can't recycle connections with aiomcache, so we need to make sure to get a new connection for every instance.
        if not self._mcache:
            self._mcache = await get_memcached_async(*args, new_connection=new_connection, **kwargs)
        return self._mcache

    async def close(self):
        if self._mcache is not None:
            log.debug("Closing AsyncIO Memcached instance %s._mcache", self.__class__.__name__)
            await self._mcache.close()
        self._mcache = None
        return await close_memcached_async()




"""
This module contains *somewhat risky* code that uses app introspection e.g. via :mod:`inspect`.

Most functions / classes in this module will **ONLY work on CPython** (the official Python interpreter from python.org),
and their functionality is not guaranteed to be stable as they interact with the interpreter to enable special functionality
such as detecting the function/class/module which called your function/method.

Functions and methods in this module may be updated with breaking API changes at any time, especially if they're needed
to adjust for a change in Python itself. Please ensure that any usage of this module is properly wrapped in a try/catch block,
and avoid relying on functions/methods in this module for critical functionality of your application.

Most useful functions:

 * :func:`.calling_function` - Returns the name of the function/method which called your function/method
 
 * :func:`.calling_module` - Returns the name of the module which called your function/method
 
 * :func:`.caller_name` - Returns the fully qualified module path to the function/method/module which called
   your function/method


**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
import inspect
from typing import Optional


def calling_function(skip=2) -> Optional[str]:
    """
    Returns the name of the function which called your function/method.
    
    Example::
        
        >>> def x(skip=2): return calling_function(skip=2)
        >>>
        >>> def y(skip=2): return x(skip)
        >>>
        >>> def z(skip=2): return y(skip)
        >>>
        >>> print(y())   # The call to x() returns that 'y' is the function which called it.
        y
        >>> print(z())   # The call to z() calls y() -> x() - still returning that 'y' is the caller of x()
        y
        >>> # If we adjust skip to 3 instead of 2, we can see that z() is the function that called y() which called x()
        >>> print(z(3))
        z
        
    :param int skip: Skip this many frames.
                     0 = calling_function()
                     1 = function which called calling_function()
                     2 = function which called the function that called calling_function() (default)
                     and so on...
     
    :return str|None function_name: Either a string containing the function name, or ``None`` if you've skipped too many frames.
    
    """
    curframe = inspect.currentframe()
    calframe = inspect.getouterframes(curframe, 2)
    
    if len(calframe) <= skip:
        return None
    return calframe[skip][3]


def last_frames():
    curframe = inspect.currentframe()
    calframe = inspect.getouterframes(curframe, 2)
    return curframe, calframe


def last_stack_frame(frame_num=2):
    return inspect.stack()[frame_num]


def calling_module(skip=2) -> Optional[str]:
    """
    Returns the name of the module which called your function/method.
    
    :param int skip: Skip this many frames.
                 0 = module containing calling_function()
                 1 = module which called calling_function()
                 2 = module which called the function that called calling_function() (default)
                 and so on...
    :return str|None mod_name: Either a string containing the module name, or ``None`` if you've skipped too many frames.
                               If called from the main python script, then ``'__main__'`` will be returned instead of a
                               proper module path.
    """
    stk = inspect.stack()
    if len(stk) <= skip:
        return None
    mod = inspect.getmodule(stk[skip][0])
    if mod is None:
        return '__main__'
    return mod.__name__


def caller_name(skip=2) -> Optional[str]:
    """
    Get the fully qualified name of a caller in the format ``some.module.SomeClass.method``
    
    .. Attention::    While class instance methods will be returned correctly, class static methods will not show up as expected.
                      The static method ``some.module.SomeClass.some_static`` would be returned as ``some.module.some_static``,
                      as if it were a top-level function in the module.
    
    
    Original source: https://stackoverflow.com/a/9812105
    
    **Basic Example**
    
    When used within the main program (the script you run ``python3 xxx.py`` on), the module will be reported as ``__main__``.
    
    File ``hello.py``::
        
        >>> from privex.helpers.black_magic import caller_name
        >>>
        >>> def f2():
        ...     return caller_name()
        >>>
        >>> def f1():
        ...     return f2()
        ...
        >>> print(f"[{__name__}] f1 result: {f1()}")
        [__main__] f1 result: __main__.f1
    
    However, as we can see below, when we create and run ``world.py`` which imports ``hello.py``, it correctly returns the
    path ``hello.f1``.
    
    File ``world.py``::
        
        >>> from hello import f1
        >>>
        >>> print(f"[{__name__}] f1 result: {f1()}")
        [hello] f1 result: hello.f1
        [__main__] f1 result: hello.f1

    
    **More Complex Example**
    
    File ``some/module/hello.py``::
        
        >>> from privex.helpers.black_magic import caller_name
        >>>
        >>> class SomeClass:
        >>>     def example_method(self, skip=2):
        ...         return caller_name(skip)
        ...
    
    File ``some/module/world.py``::
        
        >>> from some.module.hello import SomeClass
        >>>
        >>> class OtherClass:
        ...     def call_some(self, skip=2):
        ...         return SomeClass().example_method(skip)
        ...
    
    File ``test.py``::
        
        >>> from some.module.hello import SomeClass
        >>> from some.module.world import OtherClass
        >>>
        >>> def main_func():
        ...     print('SomeClass (2)', SomeClass().example_method())
        ...     print('OtherClass (1)', OtherClass().call_some(1))
        ...     print('OtherClass (2)', OtherClass().call_some())
        ...     print('OtherClass (3)', OtherClass().call_some(3))
        ...
        >>> main_func()
        SomeClass (2) test.main_func
        OtherClass (1) some.module.hello.SomeClass.example_method
        OtherClass (2) some.module.world.OtherClass.call_some
        OtherClass (3) test.main_func
    
    

    :param int skip: Specifies how many levels of stack to skip while getting caller name.
                     ``skip=1`` means "who called caller_name",
                     ``skip=2`` means "who called this function/method" etc.

    :return str caller: A fully qualified module path, e.g. ``some.module.SomeClass.some_method``
                        ``None`` is returned if skipped levels exceed stack height.
    """
    stack = inspect.stack()
    start = 0 + skip
    if len(stack) < start + 1:
        return None
    parentframe = stack[start][0]
    
    name = []
    module = inspect.getmodule(parentframe)
    
    # `modname` can be None when frame is executed directly in console
    # TODO(techtonik): consider using __main__
    if module:
        name.append(module.__name__)
    
    # detect classname
    if 'self' in parentframe.f_locals:
        # I don't know any way to detect call from the object method
        # XXX: there seems to be no way to detect static method call - it will
        #      be just a function call
        name.append(parentframe.f_locals['self'].__class__.__name__)
    codename = parentframe.f_code.co_name
    if codename != '<module>':  # top level usually
        name.append(codename)  # function or a method
    
    # Avoid circular refs and frame leaks
    # https://docs.python.org/2.7/library/inspect.html#the-interpreter-stack
    del parentframe, stack
    
    _caller = ".".join(name)
    
    if _caller in ["", None]:
        return None
    return _caller



"""
Network related helper code

**Copyright**::

        +===================================================+
        |                 © 2020 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )

"""

from privex.helpers.exceptions import BoundaryException, NetworkUnreachable

from privex.helpers.net.dns import *
from privex.helpers.net.util import *
from privex.helpers.net.common import *
from privex.helpers.net.socket import *
